//============================================================================
// MapVoteClientSplash
//============================================================================
//

class MapVoteClientSplash extends Mutator config(MapVoteXM);

#exec texture IMPORT NAME=mvxlogo File=Textures\mvxlogo.pcx GROUP=MapVoteX MIPS=Off

var MapVoteGRI MVGRI;

struct LogoNames
{
	var bool   bUse;
	var bool   bZoom;
	var bool   bTransparent;
	var string ServerLogoName;
	var string LogoSoundName;
	var int    FadeTime;
	var int    X, Y;
};

var() config bool bKillBotMesg;
var() config bool bAddBotMesg;
var() config bool bAllowSplashInGame;
var() config LogoNames SplashLogo;

var bool bInitialized;
var bool bNewClient;
var bool bSoundDone;
var bool bEnableLogo;
var bool bDisMvxLogo;
var int  iTock;
var int  GameStart;

var ChallengeHUD MyHUD;
var Texture mvx1, mvx2, ServerLogo;
var Sound MySound;
var int DrawLogo, DrawSplash, LogoCounter, SplashCounter;
var Color WhiteColor, LoadColor;
var PlayerPawn PlayerOwner;

var int TickTime;

var bool bUseLogo, bUseSplash, bDoZoom;
var bool bDone;
var bool bClear;
var string LogoName;
var string SoundName;
var string sCode;
var int    iFade;

var int Xpos, Ypos;

var FontInfo MyFonts;
var string sMessage;
var float FF, GG;

replication
{
   reliable if (Role == ROLE_Authority)
               bDone, iFade, sCode, bUseLogo, bUseSplash, bClear, LogoName, SoundName, Xpos, Ypos, GameStart,
               bEnableLogo, sMessage, bDisMvxLogo, bKillBotMesg, bAddBotMesg, bDoZoom, bAllowSplashInGame;
}

simulated function PostBeginPlay()
{
    Super.PostBeginPlay();

    if ( !bInitialized )
    {
       bInitialized = true;
       MyFonts = FontInfo(spawn(Class<Actor>(DynamicLoadObject(class'ChallengeHUD'.default.FontInfoClass, class'Class'))));
       SetTimer(0.1, true);
       FF = 0;
       GG = 0;
       bDone = False;
       SaveConfig();
    }
}

simulated function Tick( float delta )
{
    if (Level.NetMode != NM_Client && !bDone)
    {
        bUseSplash = SplashLogo.bUse;
        bDoZoom    = SplashLogo.bZoom;
        bClear     = SplashLogo.bTransParent;
        LogoName   = SplashLogo.ServerLogoName;
        SoundName  = SplashLogo.LogoSoundName;
        iFade      = SplashLogo.FadeTime;
        Xpos       = SplashLogo.X;
        Ypos       = SplashLogo.Y;
        bUseLogo   = True;
        if (iFade < 1)
            iFade = 1;
        iFade = iFade*10+510;
        bDone      = True;
    }

    if ( Level.NetMode != NM_DedicatedServer )
    {
       if ( !bHUDMutator )
          RegisterHUDMutator();
    }
}

simulated function Timer()
{
    Super.Timer();

    if ( Level.NetMode != NM_DedicatedServer )
    {
        if ( PlayerOwner != None )
        {
            if (!bNewClient)
            {
                if ( MVGRI == none )
                {
                    foreach AllActors( class'MapVoteGRI', MVGRI )
                    break;
                }

                if (TickTime++ > 20)
                {
                    bNewClient = True;
                    if (bKillBotMesg && PlayerOwner.IsA('TournamentPlayer'))
                        PlayerOwner.ClientMessage("<< Type !KillBot or !KillBots to vote 'Remove 1 or all Bots' from the Game >>");
                    if (bAddBotMesg && PlayerOwner.IsA('TournamentPlayer') && MVGRI.PlayersMax > 2)
                        PlayerOwner.ClientMessage("<< Type !AddBot or !AddBots to vote 'Add a Bot or 2 Bots' to the Game >>");
                    if (MVGRI.LoadedMapNum > 0)
                        PlayerOwner.ClientMessage("<< MVX loaded "$MVGRI.LoadedMapNum+1$" maps for this Game >>");
                    if (!MVGRI.bGameEnded)
                    {
                        if (bUseLogo && GameStart == 0 && PlayerOwner.PlayerReplicationInfo.bWaitingPlayer)
                        {
                            DrawLogo = 1;
                            SetTimer(0.10 , True);
                        }
                        if (bUseSplash)
                        {
                            if (GameStart > 0 && !bAllowSplashInGame)
                                goto NoSp;
                            DrawSplash = 1;
                            SetTimer(0.10 , True);
                        }
                    }
                }
            }
        }
NoSp:
        if ( DrawLogo > 0 && bUseLogo )
        {
            LogoCounter++;
            if ( DrawLogo == 510 )
            {
                DrawLogo = 0;
                bUseLogo = false;
            }
            else
            if ( LogoCounter > 20 )
            {
                DrawLogo += 6;
                if ( DrawLogo > 510 )
                    DrawLogo = 510;
            }
            else
            if ( LogoCounter == 60 )
            {
                DrawLogo = 5;
            }
        }

        if (iTock++ >= (iFade-510))
        {
            iTock = 510;
            if (!bUseLogo && iFade > 1)
                SetTimer(0.05, true);
            if ( DrawSplash > 0 && bUseSplash )
            {
               SplashCounter++;
               if ( DrawSplash == 510 )
               {
                   DrawSplash = 0;
                   bUseSplash = false;
               }
               else
               if ( SplashCounter > 20 )
               {
                   DrawSplash += 6;
                   if ( DrawSplash > 510 )
                       DrawSplash = 510;
               }
               else
               if ( SplashCounter == 20 )
               {
                   DrawSplash = 5;
               }
            }
        }

        if (!bUseSplash && !bUseLogo)
        {
            Disable('Tick');
            SetTimer( 0.0, False );
        }
    }
}

simulated event PostRender( Canvas C )
{
    local float Xposi, Yposi, F, ScaleX;
    local int   Xp, Yp;

    if ( PlayerOwner == none )
    {
       PlayerOwner = C.Viewport.Actor;
       MyHUD = ChallengeHUD( PlayerOwner.MyHUD );
    }

    C.Style = ERenderStyle.STY_Translucent;
    ScaleX = (1.0 * C.ClipX) / 1280.0;

    if ( !bDisMvxLogo && DrawLogo != 0 && bUseLogo && PlayerOwner.PlayerReplicationInfo.bWaitingPlayer)
    {
        if ( DrawLogo > 1 )
        {
            C.DrawColor.R = 255 - DrawLogo/2;
            C.DrawColor.G = 255 - DrawLogo/2;
            C.DrawColor.B = 255 - DrawLogo/2;
        }
        else
        {
            C.DrawColor = WhiteColor;
        }

        if (mvx1 == None)
      	    mvx1 = texture'mvxlogo';

        F = 0.95 * ScaleX;

        if (DrawLogo < 320)
            FF += 0.008 * ScaleX;
        else
            FF -= 0.008 * ScaleX;
        if (FF > 0.750 * ScaleX)
            FF = 0.750 * ScaleX;

        C.SetPos( C.ClipX - mvx1.Usize*F, C.ClipY - (mvx1.Usize * 0.95)  * ScaleX);
        C.DrawIcon( mvx1, FF );
    }

    if ( DrawSplash != 0 && bUseSplash)
    {
       if (bEnableLogo)
       {
           if (ServerLogo == None && InStr(LogoName, ".") > 0)
           {
               ServerLogo = Texture(DynamicLoadObject(LogoName, class'Texture'));
               if (MySound == None && InStr(SoundName, ".") > 0)
                   MySound = Sound(DynamicLoadObject(SoundName, class'Sound'));
           }

           if ( DrawSplash > 1 )
           {
              C.DrawColor.R = 255 - DrawSplash/2;
              C.DrawColor.G = 255 - DrawSplash/2;
              C.DrawColor.B = 255 - DrawSplash/2;
           }
           else
           {
              C.DrawColor = WhiteColor;
           }

           if (Serverlogo != None)
           {
               Xp = Xpos;
               if (Xp <= 0)
                   Xp = 50;
               Yp = Ypos;
               if (Yp <= 0)
                   Yp = 60;

               F = 0.95 * ScaleX;

               Xposi = (C.ClipX * Xp / 100) - (ServerLogo.Usize / 2);
               if (ServerLogo.USize <= 256)
                   Yposi = (C.ClipY * Yp / 100) - (ServerLogo.Vsize / 2);
               else
               if (ServerLogo.USize <= 512)
                   Yposi = (C.ClipY * Yp / 100) - (ServerLogo.Vsize / 3);
               else
                   Yposi = (C.ClipY * Yp / 100) - (ServerLogo.Vsize / 4);

               if (!bClear)
                   C.Style = ERenderStyle.STY_Normal;

               if (bDoZoom)
               {
                   if (DrawSplash < 320)
                       GG += 0.008;
                   else
                       GG -= 0.008;
                   if (GG > 1.10)
                       GG = 1.10;
               }
               else
                   GG = 1.0;

               C.SetPos( Xposi, Yposi);
               C.DrawIcon( ServerLogo, GG );

               if (!bSoundDone && MySound != None)
               {
                   bSoundDone = true;
                   PlayerOwner.PlaySound(MySound);
               }
           }
       }
    }

    ShowMessage(C);

    C.Style = ERenderStyle.STY_Normal;

    if (NextHUDMutator != None)
        NextHUDMutator.PostRender(C);
}

simulated function ShowMessage( canvas Canvas )
{
	local bool bScreen;
	local font CanvasFont;
	local Color NormalColor;

    Canvas.Style = ERenderStyle.STY_Normal;
    NormalColor = Canvas.DrawColor;
    CanvasFont = Canvas.Font;

    if (sMessage != "")
    {
       Canvas.DrawColor = LoadColor;
       bScreen = Canvas.bCenter;
       Canvas.bCenter = true;
       Canvas.Font = MyFonts.GetHugeFont(Canvas.ClipX);
       Canvas.SetPos(0 , Canvas.ClipY * 0.85);
       Canvas.DrawText(sMessage, true);
       Canvas.bCenter = bScreen;
    }

    Canvas.DrawColor = NormalColor;
	Canvas.Font = CanvasFont;
}

defaultproperties
{
    LoadColor=(R=128,G=128,B=10)
    WhiteColor=(R=255,G=255,B=255)
    bAllowSplashInGame=True
    bAlwaysRelevant=True
    RemoteRole=ROLE_SimulatedProxy
}
