//==============================================================================
// MapVote Windows ReplicationInfo - Server
//==============================================================================
//
class MapVoteWRI extends ReplicationInfo;
//
var MapVoteX MyMut;

var MapVoteTabWindow TabWindow;
var UWindowWindow TheWindow;
var() Class<UWindowWindow> WindowClass;

var() int  WinLeft;
var() int  WinTop;
var() int  WinWidth;
var() int  WinHeight;
var() bool DestroyOnClose;

var string MyData[20];      // 1=sCode,2=Version,3=Admin_Stuff,4=ServerName,5=ServerPort,6=ServerIP,
                            // 7=Stretch,8=Logoname,9=Soundname,12=mapprefix
var int    MyBool[10];      // 1=badmin,2=bwelcome,3=bwelcome,4=bShowFirst,5=imaps,6=midgamevote,7=bgameend,8=iGames

var string PlayerName[32];
var string MapVoteResults[32];
var string DefGame;
var string sTitle;
var bool   bDestroyRequested;

var int    TicksPassed;
var int    gPlayerCount;
var int    gTitleChecksum;
var int    PlayerNum;

var bool   bDebugMode;
var bool   bUpdated1;
var bool   bUpdated2;
var int    iLastCheck1;
var int    iLastCheck2;
var bool   bOpenWindowDispatched;
var bool   bSetupWindowDelayDone;

var bool   bMesgUpdate;
var bool   bAdminMesg;
var bool   bMapsOnly;
var bool   bNoBots;

var bool   bInitialized;

var int    MapsOnlyCount;
var string sCode;

var int MesgListCount, LastMesgListCount, LastMesgCount;

var string MesgList[51];

replication
{
   reliable if ( Role == ROLE_Authority && bAdminMesg)
        bAdminMesg, MesgList;

   reliable if ( Role == ROLE_Authority )
		DefGame, sCode, sTitle, bDebugMode, bMapsOnly, MapsOnlyCount, MyData, MyBool, PlayerName, MapVoteResults, UpdateMapVoteResults,
		PlayerNum, bNoBots;

   reliable if ( Role == ROLE_Authority )
        OpenWindow, CloseWindow;

   reliable if ( Role < ROLE_Authority )
        DestroyWRI;
}

event PostBeginPlay()
{
	Super.PostBeginPlay();

	if ( !bInitialized )
	{
        bUpdated1 = False;
        bUpdated2 = False;
        iLastCheck1 = 0;
        iLastCheck2 = 0;
	    OpenIfNecessary();
	    bInitialized = true;
	}
}

simulated event PostNetBeginPlay()
{
	PostBeginPlay();
	OpenIfNecessary();
}

simulated function OpenIfNecessary()
{
	local PlayerPawn P;

	if ( Owner != None )
	{
		P=PlayerPawn(Owner);
		if ( (P != None) && (P.Player != None) && (P.Player.Console != None) )
		{
			OpenWindow();
		}
	}
}

simulated function bool OpenWindow()
{
	local PlayerPawn P;
	local WindowConsole C;

	P = PlayerPawn(Owner);
	if ( P == None )
	{
		Log("#### -- Attempted to open a window on something other than a PlayerPawn");
		DestroyWRI();
		return False;
	}

	if ( P.Player.Console.bTyping )
	{
		SetTimer(0.50,False);
		return False;
	}

	bOpenWindowDispatched = True;

	C = WindowConsole(P.Player.Console);
	if ( C == None )
	{
		Log("#### -- No Console");
		DestroyWRI();
		return False;
	}

	if ( !C.bCreatedRoot || (C.Root == None) )
	{
		C.CreateRootWindow(None);
	}
	C.bQuickKeyEnable = True;
	C.LaunchUWindow();
	TicksPassed = 1;
	return True;
}

simulated function Tick (float DeltaTime)
{
	if ( TicksPassed != 0 )
	{
		if ( TicksPassed++  == 5 )
		{
			SetupWindow();
			TicksPassed=0;
		}
	}
	// 1st port of call
	if ( DestroyOnClose && (TheWindow != None) &&  !TheWindow.bWindowVisible &&  !bDestroyRequested )
	{
		bDestroyRequested=True;
		DestroyWRI();
	}
}

simulated function bool SetupWindow()
{

	if ( !bSetupWindowDelayDone )
	{
		SetTimer(1.00,False);
		return False;
	}

	if ( SetWindow() )
	{
		TabWindow = MapVoteTabWindow(MapVoteWindowFrame(TheWindow).ClientArea);
		SetTimer(1.00,False);
	}
	else
	{
		Log("Super.SetupWindow() = false");
	}
}

simulated function bool SetWindow()
{
	local WindowConsole C;

	C = WindowConsole(PlayerPawn(Owner).Player.Console);
	TheWindow = C.Root.CreateWindow(WindowClass,WinLeft,WinTop,WinWidth,WinHeight);

	if ( TheWindow == None )
	{
		Log("#### -- CreateWindow Failed");
		DestroyWRI();
		return False;
	}
	if ( C.bShowConsole )
	{
		C.HideConsole();
	}
	TheWindow.bLeaveOnscreen=True;
	TheWindow.ShowWindow();
	return True;
}

simulated function Timer ()
{
	local int i;
	local int TitleChecksum;
	local int repPlayerCount;
	local Color ColColor;

  	ColColor.R=255;
	ColColor.G=255;
	ColColor.B=10;

    if (bMesgUpdate)
    {
        MesgTimer();
        return;
    }

	if ( !bOpenWindowDispatched )
	{
		OpenWindow();
		return;
	}

	if ( !bSetupWindowDelayDone )
	{
		bSetupWindowDelayDone=True;
		SetupWindow();

        TabWindow.bNoBots = bNoBots;    //iGames                           Mydata[1]              version    bAdmin     bWelcome   bUse      bgameEnd   bShowFirst mapcount  bMidgame   Prefix
	    TabWindow.SetupMapWindow(DefGame, MyBool[8], bDebugMode, bMapsOnly, sCode, MapsOnlyCount, MyData[2], MyBool[1], MyBool[2], MyBool[3], MyBool[7], MyBool[4], MyBool[5], MyBool[6], MyData[12], PlayerNum);
        TabWindow.SetupWelcome(MyData[7], MyData[8], MyData[9], MyData[12]);
                          //   Stretch    Logoname   Soundname  Prefix
	}

    if (!bMapsOnly)       // scode     bAdmin    AdminStuff  SvrName    SvrPort    SvrIP
	    TabWindow.SetupMenu(MyData[1], MyBool[1], MyData[3], MyData[4], MyData[5], MyData[6]);

    //if ( bAdmin       || bGameEnd       || (!bGameEnd      && bMidGame) )
	if ( MyBool[1] == 1 || MyBool[7] == 1 || (MyBool[7] == 0 && MyBool[6] == 1) )
	{
        for (i=0; i<32; i++)
        {
            if (PlayerName[i] != "")
               repPlayerCount = i + 1;
        }

		if ( repPlayerCount != gPlayerCount )
		{
			for (i=gPlayerCount; i<repPlayerCount; i++)
			{
				TabWindow.AddPlayerName(PlayerName[i]); // has voted
			}
		    gPlayerCount = repPlayerCount;
		    bUpdated1=True;
		}

        for (i=0; i<31; i++)
        {
            if (MapVoteResults[i] != "")
               UpdateMapVoteResults(MapVoteResults[i], i);
        }

    	if ( TitleChecksum != gTitleChecksum )
    	{
			gTitleChecksum = TitleChecksum;
			bUpdated1=True;
		}

		if ( !bUpdated1 )
    	{
			iLastCheck1++;
			if (iLastCheck1 < 3)
			{
			    SetTimer(1.00,False);
			    return;
			}
		}
	}

    if (!bMapsOnly && bAdminMesg)
    {
        TabWindow.ClearMesgs();
        bMesgUpdate = true;
        iLastCheck2 = 0;
        MesgTimer();
    }
}

simulated function MesgTimer()
{
	local int MesgCount;
	local int MesgListCount;
	local string sTemp;
	local int i;

	for (i=1; i<51; i++)
	{
		if (MesgList[i] == "")
		   break;
		MesgListCount++;
	}

	if ( MesgListCount > LastMesgListCount )
	{

		for (i=1; i<MesgListCount+1; i++)
		{
		    sTemp = MesgList[i];
		    if (sTemp != "")
		    {
		        TabWindow.AddMessage(sTemp, i);
		    }
		}
		LastMesgListCount = MesgListCount;
		bUpdated2=True;
	}

	MesgCount = MesgListCount + MesgListCount;
	if ( MesgCount > LastMesgCount )
	{
		bUpdated2 = True;
		LastMesgCount = MesgCount;
	}

	if ( !bUpdated2 )
	{
		iLastCheck2++;
		if (iLastCheck2 < 4)
		{
		    SetTimer(1.00,False);
		    return;
		}
	}
}

simulated function UpdateMapVoteResults (string Text, int i)
{
	if ( TheWindow != None )
	{
		TabWindow.UpdateMapVoteResults(Text,i);
	}
}

simulated function CloseWindow ()
{
	local WindowConsole C;

    if (PlayerPawn(Owner) != None)     // solves access none's if player leaves
    {
	   C = WindowConsole(PlayerPawn(Owner).Player.Console);
	   C.bQuickKeyEnable = False;
    }

	if ( TheWindow != None )
	{
		TheWindow.Close();
		TabWindow.Close();
	}
}

function DestroyWRI()
{
	Destroy();
}

function DebugLog (string Msg)
{
	if (bDebugMode)
	    Log("**** MVWRI " $ Msg);
}

defaultproperties
{
    WindowClass=Class'MapVoteWindowFrame'
    WinWidth=390
    WinHeight=355
    DestroyOnClose=True
    bAlwaysRelevant=False
    RemoteRole=ROLE_SimulatedProxy
    NetPriority=2.00
}
