//============================================================================
//  MapVoteX - by ProAsm and no0ne - 2013-2022
//============================================================================

class MapVoteX extends Mutator Config(MapVoteX);

struct CustConfig
{
	var bool   bUse;
	var string MapFolder;
	var string DefaultMap;
	var string GameName;
	var string GameClass;
	var string Mutators;
	var string Settings;
	var string ServerPackages;
	var string ServerActors;
	var string ServerName;
	var bool   bNoEndGameVoting;
	var bool   bNoRandomGame;
};

struct CustURL
{
	var string Servername;
	var string IPAddress;
	var int    Port;
};

var() config bool  bAutoOpen;
var() config bool  bAllowMidGameVoting;
var() config bool  bDebugMode;
var() config bool  bMajorityPlayerVote;
var() config bool  bDoubleClickVoting;
var() config bool  bShowChatCommands;
var() config bool  bDisableMvxLogo;

var() config bool  bEnableSplashLogo;
var() config bool  bEnableWelcomeMenu;
var() config bool  bSortCustomGameList;

var() config bool  bEmptyServerSwitch;           // switch if server is empty
var() config bool  bEmptyServerSwitchMap;        // switch to default map on same game
var() config bool  bEmptyServerRandomMap;        // switch to a random map
var() config int   EmptyServerSwitchTimeMins;    // time in minutes

var() config int   GameStartVoteDelay;
var() config int   VoteTimeLimit;
var() config int   PopupWindowDelay;
var() config int   RepeatLimit;
var() config int   NoEndGameVotingDelay;

var() config float StartMesgDelay;

var() config bool  bEnableServerPackages;
var() config bool  bClearAllPackageActors;

var() config bool  bRandomBotOrder;
var() config bool  bUseServerNames;
var() config bool  bAllowPlayerKillBots;
var() config bool  bAllowPlayerAddBot;

var() config string ServerGamePassword;
var() config string ServerIpAddress;

var() config string DefaultMutators;
var() config string DefaultSettings;

var() config int    DefaultCustomGameNum;

var() config CustConfig CustomGame[200];

var() config bool   bQuickGameVote;
var() config string MVGameName;
var() config string ServerNameTitle;
var() config bool   bAddToFavorites;

Const MaxNumGames = 200;                         // max custom games
Const MaxNumMaps  = 1000;                        // max maps
Const TotalList   = 1200;

var string TotalMapList[1200];                   // max maps + max games    // xyz
var string GameList[201];
var string GameMaps[201];
var string GameNames[201];

var string Version;
var int MapCount;
var int GameCount;
var int MapsOnlyCount;

var int PlayerIDList[32];
var int PlayerVoted[32];
var int PlayerRandG[32];
var int PlayerRandM[32];
var string MapStatusText[32];

var Class<GameInfo> CustomGameClass;

var string TravelMap;
var string ServerTravelString;
var string GamePackage;
var string NewPrefix;

var bool bMidGameVote;
var bool bLevelSwitchPending;
var bool bInitMapsLoaded;
var int pos;
var int m, n;
var int MapTime, MsgPID, iSvr;
var int CurrentID;
var int CurrentPlayerCount;
var int JoinTickCount;
var int TimeLeft;
var int ScoreBoardTime;
var int ServerTravelTime;

var string LoadedGameClass;
var string LoadedGameName;
var string LoadedMapName;
var string LoadedMapPrefix;
var string UnloadMutators;
var int    LoadedGameNum;
var int    LoadedMapNum;
var int    TickRating;

var bool bInitialized1;
var bool bInitialized2;

var bool bGameHasEnded;
var bool bMidGameTimeout;

var bool bClassesSpawned;
var bool bRestarting;
var bool bMapsLoaded;
var bool bCheckRules;
var bool bMapsOnly;
var bool bGameVoted;
var bool bUTCmds;
var bool bMvxCTFGame;
var bool bNoEndGameVoting;

var string ServerName;
var string ServerPort;
var string ServerIP;
var string sCode;
var string VotedGameClass;
var string RandomMapVote;
var string RandomGameVote;

var MapVoteGRI               MVGRI;
var MapVoteRepeatLimits      MVRP;
var MapVoteClientSplash      MVCS;
var MapVoteServerPackages    MVSP;
var MapVoteServerActors      MVSA;
var MapVoteClientWelcome     MVCW;
var MapVoteWelcomeLinks      MVWL;
var MapVoteAdminMessages     MVAM;
var MapVoteCrashedMaps       MVCM;
var MapVoteMapLoader         MVML;
var MapVoteGameAliases       MVGA;
var MapVoteCommands          MVCMD;
var MapVoteIdler             MVID;
var MapVoteTopGames          MVTG;
var MapVoteTopMaps           MVTM;
var MapVoteDefaultActors     MVDA;
var MapVoteDefaultPackages   MVDF;

event Spawned()
{
    Super.Spawned();

    MVGRI = Spawn(class'MapVoteGRI');
    MVGRI.PlayersMin = DeathMatchPlus(Level.Game).MinPlayers;
    MVGRI.PlayersMax = DeathMatchPlus(Level.Game).MaxPlayers;
}

function PreBeginPlay()
{
	local int j, x;
	local MapVoteDefaultSwitch MVDS;

    Super.PreBeginPlay();

	if ( !bInitialized1 )
	{
		bInitialized1 = True;
		SaveConfig();                                      // for new installation do mvx 1st

		for (x=0; x<32; x++)
		{
			PlayerIDList[x] = -1;
		}

        bMapsOnly = false;
        bMapsLoaded = false;
        bNoEndGameVoting = false;

		CheckForGameName(MVGameName);
        LoadedGameClass = string(Level.Game.Class);
        LoadedGameName = Level.Game.GameName;

        bMvxCTFGame = false;
        if (InStr(Caps(LoadedGameClass), "CTF") >= 0)      // for goalteamscore etc
            bMvxCTFGame = true;

        GetCurrentMap();
        LoadedMapPrefix = GetMapPrefix(LoadedMapName);

        DebugLog("Starting Map = "$LoadedMapName$" | LoadedMapPrefix = "$LoadedMapPrefix);

        //Restart or execute all settings and mutators
		if (!ExecuteMutatorsAndSettings())
 		    return;                                        // game restarted

        j = Rand(900);
        if (j < 100)
            j += 100;
        sCode = string(j);

        CheckForCmds();

        MVRP = Spawn(class'MapVoteRepeatLimits');

        MVCMD = Spawn(class'MapVoteCommands');
        MVCMD.bShowChatCommands = bShowChatCommands;
        MVCMD.bKillBots = bAllowPlayerKillBots;
        MVCMD.bAddBot   = bAllowPlayerAddBot;
        MVCMD.bUTCmds = bUTCmds;
        MVCMD.myMut = self;

		CustomGameClass = None;
		SaveConfig();

        DeathMatchPlus(Level.Game).NetWait = 15;

        MVTG = Spawn(class'MapVoteTopGames');
        MVTG.myMut = self;
        MVTG.MVGRI = MVGRI;
        MVTG.LoadStuff();

        MVTM = Spawn(class'MapVoteTopMaps');
        MVTM.myMut = self;
        MVTM.MVGRI = MVGRI;
        MVTM.LoadStuff();

        if (bEnableWelcomeMenu)
        {
           MVCW = Spawn(class'MapVoteClientWelcome');
           MVCW.MyMut = self;
        }

        if (EmptyServerSwitchTimeMins < 3)
        {
            EmptyServerSwitchTimeMins = 3;
            SaveConfig();
		}

        if (bEmptyServerSwitch)
        {
           MVDS = Spawn(class'MapVoteDefaultSwitch');
           MVDS.MyMut = self;
           MVDS.ServerSwitchSecs = EmptyServerSwitchTimeMins * 60;
        }
	}
}

function PostBeginPlay ()
{
    Super.PostBeginPlay();

    if ( !bInitialized2 )
    {
        bInitialized2 = true;

        MVCS = Spawn(class'MapVoteClientSplash');
        MVCS.bEnableLogo = bEnableSplashLogo;
        MVCS.bDisMvxLogo = bDisableMvxLogo;
        MVCS.sCode = sCode;

        if (MVRP != None)
        {
            MVRP.RepeatLimit = RepeatLimit;
            MVRP.AdjustMapRepeats(LoadedGameNum);
            MVRP.MyMut = self;
        }

        if (MVGA == None)
            MVGA = Spawn(class'MapVoteGameAliases');

        if (bEnableServerPackages && class'MapVoteDefaultPackages'.default.DefaultPackage[0] == "")
        {
            Spawn(class'MapVoteDefaultPackages');          // only happens on 1st installation
            Spawn(class'MapVoteDefaultActors');
        }

        if (MVID == None)
            MVID = Spawn(class'MapVoteIdler');

        if (Level.Game.IsA('CoopGame'))
            MVGRI.bCoopGame = True;

        CheckTickRate(0, True);
        GetServerInfo();
        MsgPID = -1;
    }
}

function CheckForCmds()
{
    local string sTemp;
    local Mutator M;
    local int i;

    bUTCmds = false;
    for (M = Level.Game.BaseMutator; M != None; M = M.NextMutator)
    {
        sTemp = sTemp$","$M.Class;
    }

    DebugLog("MapVoteX: Mutator List = "$sTemp);

    i = InStr(Caps(sTemp), "UTCMDS");
    if (i != -1)
        bUTCmds = true;
}

function GetServerInfo()
{
    local int i;

    ServerIP = ServerIpAddress;
    if (Len(ServerIP) < 7)
    {
        ServerIP = Level.GetAddressURL();  // ServerIP = 0.0.0.0:7777 | Port = 7777
    }
    if (ServerIP == "" || Left(ServerIP, 1) == "0")
        return;

    i = InStr(ServerIP, ":");
    if (i >= 0)
    {
        ServerPort = Mid(ServerIP, i+1);
        ServerIP = Left(ServerIP, i);
        i = int(ServerPort);
        i++;
        ServerPort = string(i);
    }
    else
    {
        i = Level.Game.GetServerPort();    // 7777
        i += 1;                            // 7778
        ServerPort = string(i);
    }

    if (ServerName == "")
        ServerName = Level.Game.GameReplicationInfo.ServerName;
}

function SetServername(string sName)
{
    if (bUseServerNames)
    {
        if (Level.Game.GameReplicationInfo != None)
        {
            Level.Game.GameReplicationInfo.ServerName = sName;
            Level.Game.GameReplicationInfo.ShortName  = sName;
            Level.Game.GameReplicationInfo.static.StaticSaveConfig();
        }

        class'Engine.GameReplicationInfo'.default.ServerName = sName;
        class'Engine.GameReplicationInfo'.default.ShortName  = sName;
        class'Engine.GameReplicationInfo'.static.StaticSaveConfig();
    }
}

function CheckServername()
{
    local string sTemp;

    if (!bUseServerNames)
        return;

    if (Level.Game.GameReplicationInfo == None)
        return;

    if (LoadedGameNum < 0)
        return;

    sTemp = CustomGame[LoadedGameNum].ServerName;
    if (ServerNameTitle != "")
        sTemp = ServerNameTitle$sTemp;
    if ( sTemp != "")
    {
        if (Level.Game.GameReplicationInfo.ServerName != sTemp || Level.Game.GameReplicationInfo.ShortName != sTemp)
            SetServername(sTemp);
        MVGRI.SvrName = sTemp;
    }

    if (CustomGame[LoadedGameNum].bNoEndGameVoting)
        bNoEndGameVoting = True;
}

function bool HandleEndGame ()
{
	Super.HandleEndGame();

	if (CheckForTie())
	{
		return False;
	}

	if (Level.Game.IsA('Assault'))
	{
		if (!Assault(Level.Game).bDefenseSet)
		{
			return False;
		}
	}

	bMidGameTimeout = false;
	bMidGameVote = false;
	CloseAllVoteWindows();
	ClearAllVoting();

    DeathMatchPlus(Level.Game).bDontRestart = True;
    if (bNoEndGameVoting)
    {
        TimeLeft = NoEndGameVotingDelay;
        ScoreBoardTime = 2;
    }
    else
    {
        TimeLeft = VoteTimeLimit;
        ScoreBoardTime = PopupWindowDelay;
    }
   	SetTimer(1.00,True);
    return False;
}

function GetCurrentMap()
{
   local int i;

   i = InStr(Caps(Level.Game), ".");

   if (i != -1)
   {
      LoadedMapName = Left(Level.Game, i);
   }
   else
   {
      i = InStr(Caps(Level.Game.BaseMutator), ".");
      if (i != -1)
      {
         LoadedMapName = Left(Level.Game.BaseMutator, i);
      }
   }
}

function CheckForGameName(string sName)
{
    local string sTemp;
    local int i;

    LoadedGameNum = -1;

    for (i=0; i<MaxNumGames; i++)
    {
        if (CustomGame[i].bUse && CustomGame[i].GameName ~= sName)
        {
           LoadedGameNum = i;
           if (CustomGame[i].GameClass ~= string(Level.Game.Class))
               Level.Game.GameName = CustomGame[i].GameName;
           if (bUseServerNames && CustomGame[i].Servername != "")
           {
               sTemp = CustomGame[i].ServerName;
               if (ServerNameTitle != "")
                   sTemp = ServerNameTitle$sTemp;
               SetServername(sTemp);
               ServerName = sTemp;
           }
           DebugLog("Searching for Game: "$sName$" and found: "$CustomGame[i].GameName);
           break;
        }
    }
    for (i=0; i<MaxNumGames; i++) // used in maplist and serveractorpackages
    {
        default.GameNames[i] = CustomGame[i].GameName;
    }
}

function CheckForGameClass(string sClass)
{
    local string sTemp;
    local int i;

    LoadedGameNum = -1;

    for (i=0; i<MaxNumGames; i++)
    {
        if (CustomGame[i].bUse && CustomGame[i].GameClass ~= sClass)
        {
           LoadedGameNum = i;
           Level.Game.GameName = CustomGame[i].GameName;
           sTemp = CustomGame[i].ServerName;
               if (ServerNameTitle != "")
                   sTemp = ServerNameTitle$sTemp;
           if (bUseServerNames && sTemp != "")
           {
               SetServername(sTemp);
               ServerName = sTemp;
           }
           DebugLog("Searching for Game: "$sClass$" and found: "$CustomGame[i].GameClass);
           break;
        }
    }
}

function SetMessage(bool bSet)
{
    if (MVCS == None)
        MVCS  = Spawn(class'MapVoteClientSplash');

    MVCS.sMessage = "";

    if (bSet)
    {
        MVCS.sMessage = "PRELOADING MAPS";       // normal large text
        return;
    }
}

event Timer()
{
	local Pawn P, aPawn;
	local int VoterNum;
	local int NoVoteCount;
	local int MapNum;

    if (iSvr++ >= 300)                           // 5 mins
    {
        iSvr = 0;
        CheckServername();
    }

	if ( bLevelSwitchPending )
	{
		if ( Level.TimeSeconds > ServerTravelTime )
		{
			if ( Level.NextURL == "" )
			{
				if ( Level.NextSwitchCountdown < 0 )
				{
					HandleFailedMap(TravelMap);
					MapNum = GetRandomMap();
					ServerTravelString = SetupGameMap(TotalMapList[MapNum]);
				}
				Level.ServerTravel(ServerTravelString,False);
			}
		}
		return;
	}

	if ( ScoreBoardTime > 0 )
	{
		ScoreBoardTime--;
		if ( ScoreBoardTime == 0 )
		{
			if (Level.Game.bGameEnded && bNoEndGameVoting && !bMapsOnly)
			    return;

            if (Level.Game.bGameEnded && LoadedMapNum == 1 && !bMapsOnly)
            {
                Level.ServerTravel("?restart", false);
                return;
            }

			if (bMapsOnly)
			{
			    if (!bMapsLoaded)
			    {
			        ScoreBoardTime = 5;
			        return;
			    }
			    if (bQuickGameVote)
			        CloseAllVoteWindows();
			    DebugLog("Timer - Opening All Windows for New Game Maps.");
			    TimeLeft = VoteTimeLimit;
			}

			for ( aPawn = Level.PawnList; aPawn != None; aPawn = aPawn.nextPawn )
			{
				if ( (PlayerPawn(aPawn) != None) && aPawn.bIsPlayer )
				{
					VoterNum = FindPlayerIndex(PlayerPawn(aPawn).PlayerReplicationInfo.PlayerID,True);
					if ( VoterNum > -1 )
					{
						if ( PlayerVoted[VoterNum] == 0 )
						{
                            if (Level.Game.bGameEnded && !bAutoOpen)
                                continue;
							OpenVoteWindow(PlayerPawn(aPawn),False,0);  // timer open all
						}
					}
				}
			}
			BroadcastMessage(string(TimeLeft) $ " seconds left to vote.", True);
		}
		return;
	}

    if (TimeLeft >= 0)
	    TimeLeft--;

    if (Level.Game.bGameEnded && bNoEndGameVoting && !bMapsOnly)
        goto Timerz;

	if ( TimeLeft == 60 )
	{
		BroadcastMessage("1 Minute left to vote.", True);

		for ( P=Level.PawnList; P != None; P=P.nextPawn )
		{
			if ( P.IsA('TournamentPlayer') )
			{
				TournamentPlayer(P).TimeMessage(12);
			}
		}
	}

	if ( TimeLeft == 10 )
	{
		BroadcastMessage("10 seconds left to vote.", True);
	}

Timerz:

	if ((TimeLeft < 11) && (TimeLeft > 0))
	{
		if (Level.Game.bGameEnded)
		{
		   for (P=Level.PawnList; P != None; P=P.nextPawn)
		   {
			   if (P.IsA('TournamentPlayer'))
			   {
				   TournamentPlayer(P).TimeMessage(TimeLeft);
			   }
		   }
		}
		else
		{
		   if (TimeLeft < 3)
		   {
		       bMidGameTimeout = true;
		   }
		}
	}

	CleanUpPlayerIDs();

    if (TimeLeft % 20 == 0 && TimeLeft > 0)
    {
        NoVoteCount = 0;
        // force all players voting windows open if they have not voted
        for ( aPawn=Level.PawnList; aPawn!=None; aPawn=aPawn.NextPawn )
        {
            if (aPawn.bIsPlayer && PlayerPawn(aPawn) != none)
            {
                VoterNum = FindPlayerIndex(PlayerPawn(aPawn).PlayerReplicationInfo.PlayerID, True);
                if (PlayerVoted[VoterNum] == 0) // if this player has not voted
                {
                    NoVoteCount++;
                }
            }
        }

		if ( NoVoteCount == 0 )
		{
			TallyVotes(True);
		}
	}

	if ( TimeLeft == 0 )
	{
		TallyVotes(True);
	}
}

function Tick(float DeltaTime)
{
	local string sTemp;
	local Pawn P, Other;

	Super.Tick(DeltaTime);

    if (LoadedMapName ~= "DM-Tutorial")
        return;

	if ( !bInitMapsLoaded )
	{
		bInitMapsLoaded=True;
		GotoState('Loading');
	}

	if ( Level.Game.NumPlayers > CurrentPlayerCount )
	{
		CurrentPlayerCount = Level.Game.NumPlayers;
	}

	if ( Level.Game.NumPlayers < CurrentPlayerCount )
	{
		CurrentPlayerCount = Level.Game.NumPlayers;

		DebugLog("CurrentPlayerCount decreased:" @ string(CurrentPlayerCount));

		CleanUpPlayerIDs();
		if ( Level.Game.NumPlayers > 0 )
		{
		    if (Level.Game.NumPlayers == 1 && TimeLeft > 1 && !bGameHasEnded)
		    {
		        TimeLeft = 0;
		        bMidGameTimeout = true;
		        TallyVotes(True);
		    }
		    else
		    {
			    TallyVotes(False);
			    UpdateOpenWRI(False);
			}
		}
	}

	if ( Level.Game.CurrentID > CurrentID )
	{
		if ( JoinTickCount < 20 )
		{
			JoinTickCount++;
			return;
		}
		JoinTickCount=0;

        // Find the new player
        for ( Other = Level.PawnList; Other != None; Other = Other.NextPawn )
        {
            if (Other.IsA('PlayerPawn') && (Other.PlayerReplicationInfo != None) && (Other.PlayerReplicationInfo.PlayerID == CurrentID) )
               break;
        }

		CurrentID++;
        CheckServername();

		if ( (Other == None) ||  !Other.bIsPlayer ||  !Other.IsA('PlayerPawn') || Other.PlayerReplicationInfo == None)
		{
			return;
		}

		if ( Level.NetMode != NM_Standalone )
		{
		    sTemp = PlayerPawn(Other).GetPlayerNetworkAddress();
		    sTemp = Left(sTemp, InStr(sTemp, ":"));
		    Log("MapVoteX: New Player = "$Other.PlayerReplicationInfo.PlayerName$" | IP = "$sTemp$" | Time = "$GetAbsoluteTime());
		}
	}

	if (Level.Game.bGameEnded && !bGameHasEnded)
	{
	    bGameHasEnded = true;
	    MVGRI.bGameEnded = true;
    }

    if (bDisableMvxLogo)
    {
	    for (P = Level.PawnList; P != None; P = P.NextPawn)
	    {
	        if (P.IsA('PlayerPawn') && P.bIsPlayer && P.PlayerReplicationInfo != None)
	        {
	            if (Level.TimeSeconds - P.PlayerReplicationInfo.StartTime >= StartMesgDelay && P.PlayerReplicationInfo.PlayerID > MsgPID)
	            {
	                CheckForCmds();
                    P.ClientMessage(">>> Running MapVoteX "$Version$" <<<");
	                if (bUTCmds)
                        P.ClientMessage(">>> Type !MVX into chat to open Mapvote Window." );
                    else
                        P.ClientMessage(">>> Mutate MapvoteX VoteMenu to open Mapvote Window." );
	                MsgPID = P.PlayerReplicationInfo.PlayerID; // Increase to keep track of whom still to message
	            }
	        }
	    }
	}
}

function ModifyPlayer( Pawn Other )
{
    Super.ModifyPlayer( Other );

    if (MVCS != None)
        MVCS.GameStart = 1;
    if (!bCheckRules)
        CheckRules();
}

function CheckRules()
{
    local int i;

    Level.Game.GetRules();
    if (Level.Game.EnabledMutators != "")
    {
        i = InStr(Level.Game.EnabledMutators, "MapVoteX108");
        if (i == -1)
            Level.Game.EnabledMutators = Level.Game.EnabledMutators$", MapVoteX108";
        bCheckRules = true;
    }
}

function HandleFailedMap(string MapName)
{
	MapName=GetPureMapname(MapName);
	Log("Map change Failed, bad or missing map file - MapName:" @ MapName);
    if (MVAM == None)
        MVAM = Spawn(class'MapVoteAdminMessages');
    MVAM.SaveAdminMessage("MVX", "The map: "$Mapname$" has crashed due to failure or could not be found!");
    if (MVCM == None)
        MVCM = Spawn(class'MapVoteCrashedMaps');
    MVCM.StoreCrashedMap(MapName);
}

function bool CheckForTie ()
{
	local TeamInfo Best;
	local int i;
	local Pawn P;
	local Pawn BestP;

	if ( Level.Game.IsA('Assault') )
	{
		return False;
	}
	if ( Level.Game.IsA('Domination') )
	{
		return False;
	}

	if ( Level.Game.IsA('TeamGamePlus') )
	{   // find best team
		for (i=0; i<TeamGamePlus(Level.Game).MaxTeams; i++)
		{
			if ( (Best == None) || (Best.Score < TeamGamePlus(Level.Game).Teams[i].Score) )
			{
				Best = TeamGamePlus(Level.Game).Teams[i];
			}
		}

		for (i=0; i<TeamGamePlus(Level.Game).MaxTeams; i++)
		{
			if ( (Best.TeamIndex != i) && (Best.Score == TeamGamePlus(Level.Game).Teams[i].Score) )
			{
				return True; // teams tied
			}
		}
	}
	else  // check for death match tie
	{
        // find individual winner
        for (P = Level.PawnList; P != None; P = P.nextPawn)
        {
            if (P.bIsPlayer && ((BestP == None) || (P.PlayerReplicationInfo.Score > BestP.PlayerReplicationInfo.Score)) )
               BestP = P;
        }
        // check for tie
        for (P = Level.PawnList; P != None; P = P.nextPawn )
        {
            if (P.bIsPlayer && (BestP != P) && (P.PlayerReplicationInfo.Score == BestP.PlayerReplicationInfo.Score) )
               return true; // tied
        }
	}
	return False;
}

function SubmitVote(string MapName, PlayerPawn Voter) // gets full mapname including [RR] and [RG]
{
	local int i, PlayerIndex, MapIndex;
	local bool bRandM, bRandG, bFound;
	local string sMap;

	if ( Voter.IsA('Spectator') && !CheckForAdmin(Voter) )
	{
		Voter.ClientMessage("Spectators are not allowed to vote.");
		return;
	}

	if ( bLevelSwitchPending )
	{
		return;
	}

	if ( Left(MapName,4) == "[XX]" )
	{
		return;
	}

    if ( Left(MapName,4) == "[RR]" )
    {
        for (i=0; i<32; i++)                               // 1st check if someone has already voted random
        {
             if (PlayerRandM[i] == 1)
             {
                 MapName = TotalMapList[PlayerVoted[i]];
                 bFound = true;
                 break;
             }
        }
        if (!bFound)                                       // no one has yet voted random
        {
            MapName = GetPureMapname(MapName);             // gets new random map
            RandomMapVote = Mapname;                       // for pri reference;
	        Log(Voter.PlayerReplicationInfo.PlayerName$" has voted for a Random Map");
	    }
	}
    else
    if ( Left(MapName,4) == "[RG]" )
    {
        for (i=0; i<32; i++)                               // 1st check if someone has already voted random
        {
             if (PlayerRandG[i] == 1)
             {
                 MapName = TotalMapList[PlayerVoted[i]]; //MapName = GameList[PlayerVoted[i]];
                 bFound = true;
                 break;
             }
        }
        if (!bFound)                                       // no one has yet voted random
        {
            MapName = GetPureMapname(MapName);             // gets new random game
            RandomGameVote = Mapname;                          // for pri reference;
	        Log(Voter.PlayerReplicationInfo.PlayerName$" has voted for a Random Game");
	    }
	}
    else
    {
        Log(Voter.PlayerReplicationInfo.PlayerName $ " has voted for: " $ MapName); //Jojoza has voted for Capture the Flag
    }

	CleanUpPlayerIDs();
	PlayerIndex = FindPlayerIndex(Voter.PlayerReplicationInfo.PlayerID,True);

	if ( PlayerIndex == -1 )
	{
		return;
	}

    MapIndex = GetMapIndex(MapName);

    if ( MapIndex == 0 )                 // problem
    {
        return;
	}

	if ( PlayerVoted[PlayerIndex] ~= MapIndex )
	{
	    return;
	}

    if (Left(MapName, 4) == "[GG]")
    {
        if (Mid(MapName, 4) ~= RandomGameVote)
            bRandG = true;
    }

    if (MapName ~= RandomGameVote)
        bRandG = true;

    if (MapName ~= RandomMapVote)
        bRandM = true;

	PlayerVoted[PlayerIndex] = MapIndex;

	if (bRandM)
	    PlayerRandM[PlayerIndex] = 1;

	if (bRandG)
	    PlayerRandG[PlayerIndex] = 1;

	if (MapName ~= LoadedMapName && !bRandM)
	    BroadcastMessage(Voter.PlayerReplicationInfo.PlayerName $ " voted for a Map Restart", True);
	else
	if (bRandM)
	    BroadcastMessage(Voter.PlayerReplicationInfo.PlayerName $ " voted for a Random Map", True);
	else
	if (bRandG)
	    BroadcastMessage(Voter.PlayerReplicationInfo.PlayerName $ " voted for a Random Game", True);
	else
	{
	    sMap = MapName;
        if (Left(MapName, 4) == "[GG]")
            sMap = Mid(MapName, 4);
        BroadcastMessage(Voter.PlayerReplicationInfo.PlayerName $ " voted for " $ sMap, True);
    }
    MVGRI.bVoting = True;
    TallyVotes(False);
}

// Sort the ini customgame list
function SortIniGames()
{
    local int a, b;
    local string sG1, sG2, TempName, DefGame;
    local CustConfig CustGames[MaxNumGames];
    local string GNames[MaxNumGames];

    for (a=0; a<MaxNumGames; a++)
    {
        if (CustomGame[a].GameName == "")
            GNames[a] = "~~~~~~";
        else
            GNames[a] = CustomGame[a].GameName;
    }

    DefGame = CustomGame[DefaultCustomGameNum].GameName;

    // bubble sort the game list
    for (a=0; a<=MaxNumGames-1; a++)
    {
       for (b=a+1; b<=MaxNumGames; b++)
       {
           if ((GNames[a] != "" && GNames[b] != "") && (Caps(GNames[a]) > Caps(GNames[b])))
           {
               TempName  = GNames[a];
               GNames[a] = GNames[b];
               GNames[b] = TempName;
           }
       }
    }

    for (a=0; a<MaxNumGames; a++)
    {
        if (GNames[a] == DefGame)
        {
            DefaultCustomGameNum = a;
            break;
        }
    }

    for (a=0; a<MaxNumGames; a++)
    {
        for (b=0; b<MaxNumGames; b++)
        {
            sG1 = GNames[a];
            sG2 = CustomGame[b].GameName;
            if (sG1 == sG2)
            {
                CustGames[a].bUse           = CustomGame[b].bUse;
                CustGames[a].MapFolder      = CustomGame[b].MapFolder;
                CustGames[a].DefaultMap     = CustomGame[b].DefaultMap;
                CustGames[a].GameName       = CustomGame[b].GameName;
                CustGames[a].GameClass      = CustomGame[b].GameClass;
                CustGames[a].Mutators       = CustomGame[b].Mutators;
                CustGames[a].Settings       = CustomGame[b].Settings;
                CustGames[a].ServerPackages = CustomGame[b].ServerPackages;
                CustGames[a].ServerActors   = CustomGame[b].ServerActors;
                CustGames[a].ServerName     = CustomGame[b].ServerName;
                CustGames[a].bNoEndGameVoting = CustomGame[b].bNoEndGameVoting;
                CustGames[a].bNoRandomGame  = CustomGame[b].bNoRandomGame;
            }
        }
    }

    for (a=0; a<MaxNumGames; a++)
    {
        CustomGame[a].bUse           = CustGames[a].bUse;
        CustomGame[a].MapFolder      = CustGames[a].MapFolder;
        CustomGame[a].DefaultMap     = CustGames[a].DefaultMap;
        CustomGame[a].GameName       = CustGames[a].GameName;
        CustomGame[a].GameClass      = CustGames[a].GameClass;
        CustomGame[a].Mutators       = CustGames[a].Mutators;
        CustomGame[a].Settings       = CustGames[a].Settings;
        CustomGame[a].ServerPackages = CustGames[a].ServerPackages;
        CustomGame[a].ServerActors   = CustGames[a].ServerActors;
        CustomGame[a].ServerName     = CustGames[a].ServerName;
        CustomGame[a].bNoEndGameVoting = CustGames[a].bNoEndGameVoting;
        CustomGame[a].bNoRandomGame  = CustGames[a].bNoRandomGame;
    }

    SaveConfig();
    Level.ServerTravel("?restart", false);
}

function bool CheckForAdmin(PlayerPawn aPlayer)
{
    if ( aPlayer.PlayerReplicationInfo.bAdmin || aPlayer.bAdmin )
        return true;
    else
        return false;
}

function Mutate (string MutateString, PlayerPawn Sender)
{
    local string MapName, GamName, sTemp, sNick;
	local int i;

    DebugLog("MutateString = "$MutateString);

	if ( Left(Caps(MutateString),10) == "BDBMAPVOTE" )
	     MutateString = "MAPVOTEX VOTEMENU";

    if (Level.Game == None)
        return;

    if (Level.Game.bGameEnded && bNoEndGameVoting && !CheckForAdmin(Sender) && !bMapsOnly)
    {
        if (Caps(MutateString) == "MAPVOTEX VOTEMENU")
            Sender.ClientMessage("Voting has been disabled!");
        goto Mutatez;
    }

	if ( Left(Caps(MutateString),8) == "MAPVOTEX" )
	{
        if ((!bMapsLoaded) || (ScoreBoardTime > 1))
        {
            if (!CheckForAdmin(Sender))
            {
                Sender.ClientMessage("Please wait while the maps are being loaded.");
                return;
            }
        }
        MutateString = Mid(MutateString, 9);
        sNick = Sender.PlayerReplicationInfo.PlayerName;
		if ( Caps(MutateString) == "VOTEMENU" )
		{
			if ((Level.TimeSeconds > GameStartVoteDelay) || (Level.NetMode == 0) || (CheckForAdmin(Sender)))
			{
				CleanUpPlayerIDs();
				DebugLog("TimeLimit = "$DeathMatchPlus(Level.Game).TimeLimit*60$" | RemainingTime = "$DeathMatchPlus(Level.Game).RemainingTime$" | Level.TimeSeconds = "$Level.TimeSeconds$" | GameStartVoteDelay = "$GameStartVoteDelay);
				OpenVoteWindow(Sender,True,1);
			}
			else
			{
				Sender.ClientMessage("Please wait "$GameStartVoteDelay$" seconds to vote");
			}
			return;
		}

        if (MutateString ~= "SpawnFiles" && CheckForAdmin(Sender))
        {
            SpawnAdvancedClasses();
            Sender.ClientMessage("All Advanced File Classes have been Respawned!");
            return;
        }

        if ( MutateString ~= "ClearPackages" || MutateString ~= "ClearActors" )
        {
            if (CheckForAdmin(Sender))
            {
                Sender.ClientMessage("Server Packages and Actors will be cleared next level.");
                bClearAllPackageActors = True;
                SaveConfig();
            }
            else
                Sender.ClientMessage("You need admin access for this option.");
        }

        if (Left(MutateString, 3) != sCode)
            return;

        MutateString = Mid(MutateString, 3);
		if ( Left(MutateString, 3) ~= "MAP" )
		{
			MapName = Mid(MutateString, 4);
			if (CheckForAdmin(Sender))
			{
				DebugLog("Admin Forced Map TravelString: "$ServerTravelString);
                MapName = GetPureMapname(MapName);
				BroadcastMessage("Server Admin has force a map switch to " $ MapName, True);
				CloseAllVoteWindows();
				if (MapName ~= LoadedMapName)
				{
				    Level.ServerTravel("?restart", false);
				    return;
				}
				ServerTravelString = SetupGameMap(MapName);
				ServerTravelTime = Level.TimeSeconds + 2;
		        bLevelSwitchPending = True;
		        SetTimer(1.0, true);
			}
			else
			{
				SubmitVote(MapName, Sender);
			}
			return;
		}

		if ( Left(MutateString, 4) ~= "GAME" )    // GAME [RG]Random
		{
			GamName=Mid(MutateString, 5);         // [RG]Random
			if (CheckForAdmin(Sender))
			{
				if (Left(GamName, 4) == "[RG]")   //[RG]Random
				{
				    GamName = "[GG]"$GetPureMapname(GamName);
				    BroadcastMessage("Server Admin has force a switch to Random Game", True);
				}
				else
				    BroadcastMessage("Server Admin has force a switch to " $Mid(GamName, 4), True);

				DebugLog("Server Admin has force a switch to "$Mid(GamName, 4));
				ServerTravelString = SetupGameMap(GamName);
				ServerTravelString = ServerTravelString$GetInlineSettings(0, GamName);
				CloseAllVoteWindows();
				ServerTravelTime = Level.TimeSeconds + 2;
		        bLevelSwitchPending = True;
		        SetTimer(1.0, true);
		        DebugLog("Admin Forced Game TravelString: "$ServerTravelString);
			}
			else
			{
				if (Mid(GamName, 4) ~= MVGameName)
				{
				   Sender.ClientMessage("You cannot vote for the current game! ");
				   return;
				}
				SubmitVote(GamName, Sender);
			}
			return;
		}

        if ( Left(MutateString, 12) ~= "AdminMessage" )
        {

             Sender.ClientMessage("Message for Admin Received." );
             if (MVAM == None)
                 MVAM = Spawn(class'MapVoteAdminMessages');
             MVAM.SaveAdminMessage(Sender.PlayerReplicationInfo.PlayerName, Mid(MutateString, 13));
             return;
        }

        if (CheckForAdmin(Sender))
        {
            if ( MutateString ~= "ForceGameEnd" )
            {
                BroadcastMessage( Sender.PlayerReplicationInfo.PlayerName @ "forced the game to end." );
                Level.Game.EndGame( "forced by admin" );
            }
            else
            if ( MutateString ~= "PauseGame" )
            {
                if (Level.Pauser == "")
                {
 		            Level.Pauser = Sender.PlayerReplicationInfo.PlayerName;
		            BroadCastMessage("Game paused by Admin.");
                }
                else
                {
		            Level.Pauser = "";
		            BroadCastMessage("Game restarted by Admin.");
                }
            }
            else
            if ( MutateString ~= "AddBot" )
            {
                if (DeathMatchPlus(Level.Game).NumBots < 16)
                {
                   Level.Game.ForceAddBot();
                   BroadcastMessage("Admin added a Bot to the game.");
                   return;
                }
                Sender.ClientMessage("Bots in game are already at maximum.");
                return;
            }
            else
            if ( MutateString ~= "RemBot" )
            {
               if (RemoveBot(False))
                   BroadcastMessage("Admin removed a Bot from the game.");

            }
            else
            if ( Left(MutateString, 16) ~= "AdminDefaultGame" )
            {
               MutateString = Mid(MutateString, 17);
               i = InStr(MutateString, ",");
               sTemp = Left(MutateString, i);    // bUse switch
               GamName = Mid(MutateString, i+1);

               if (sTemp != "None")
               {
                   for (i=0; i<MaxNumGames; i++)
                   {
                       if (CustomGame[i].GameName ~= sTemp)
                       {
                           CustomGame[i].bUse = !CustomGame[i].bUse;
                           SaveConfig();
                           UpdateGRI();
                           Sender.ClientMessage("The Game: "$sTemp$" has been set as:  Active = "$CustomGame[i].bUse);
                           break;
                       }
                   }
               }

               if (GamName != "None" && GamName != CustomGame[DefaultCustomGameNum].GameName)
               {
                   for (i=0; i<MaxNumGames; i++)
                   {
                       if (CustomGame[i].GameName ~= GamName)
                       {
                           if (CustomGame[i].bUse == false)
                           {
                               Sender.ClientMessage("ERROR! "$GamName$" has been Disabled - bUse=False");
                               return;
                           }
                           Sender.ClientMessage(Gamname$" has been set as the Default Game.");
                           DefaultCustomGameNum = i;
                           SaveConfig();
                           HandleDefaultSwitch(True);
                           return;
                       }
                   }
               }
            }
            else
            if ( MutateString ~= "DeleteAdminMessages" )
            {
                if (MVAM == None)
                    MVAM = Spawn(class'MapVoteAdminMessages');
                MVAM.DeleteMessages();
                Sender.ClientMessage("All Admin Messages have been deleted.");
                return;
            }
            else
            if ( Left(MutateString, 15) ~= "SortCustomGames" )
            {
                SortIniGames();
                Sender.ClientMessage("CustomGames sorted - Server will restart..");
                return;
            }
            else
            if ( MutateString ~= "DeleteTopGames" )
            {
                if (MVTG == None)
                    MVTG = Spawn(class'MapVoteTopGames');
                MVTG.DeleteGames();
                Sender.ClientMessage("The bottom half of the TopGames have been deleted.");
                return;
            }
            else
            if ( MutateString ~= "DeleteTopMaps" )
            {
                if (MVTM == None)
                    MVTM = Spawn(class'MapVoteTopMaps');
                MVTM.DeleteMaps();
                Sender.ClientMessage("The bottom half of the TopMaps have been deleted.");
                return;
            }
            else
            if ( Left(MutateString, 5) ~= "Admin" )
            {
                SetAdminStuff(Mid(MutateString, 6));
                Sender.ClientMessage("Game Configurations have been updated.");
            }
            else
            if ( MutateString == "RestartServer" )
            {
                Sender.ClientMessage("Restarting Level.");
                SaveConfig();
                Level.ServerTravel("?restart", false);
            }
        }
        return;
	}

Mutatez:
	if (NextMutator != None)
		NextMutator.Mutate(MutateString, Sender);
}

//autoopen,midgamevoting,debugmode,!cmds,bmajority,logonsplash,advertising,bwelcomefirst,sortgamelist,bEmptyServerSwitchMap,bEmptyServerRandomMap,
//emptyservertime,startvotedelay,votetimelimit,menupopup,repeatlimit
function SetAdminStuff(string Stuff)
{
    local int i;
    local string sTemp;

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bAutoOpen = bool(i);
    default.bAutoOpen = bool(i);

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bAllowMidGameVoting = bool(i);
    default.bAllowMidGameVoting = bool(i);

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bDebugMode = bool(i);
    default.bDebugMode = bool(i);

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bAllowPlayerKillBots = bool(i);
    default.bAllowPlayerKillBots = bool(i);

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bAllowPlayerAddBot = bool(i);
    default.bAllowPlayerAddBot = bool(i);

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bMajorityPlayerVote = bool(i);
    default.bMajorityPlayerVote = bool(i);

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    MVID.bUseSendToSpec = bool(i);
    MVID.SaveConfig();

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bEnableWelcomeMenu = bool(i);
    default.bEnableWelcomeMenu = bool(i);

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bSortCustomGameList = bool(i);
    default.bSortCustomGameList = bool(i);

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bEmptyServerSwitch = bool(i);
    default.bEmptyServerSwitch = bool(i);

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bEmptyServerRandomMap = bool(i);
    default.bEmptyServerRandomMap = bool(i);

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    bNoEndGameVoting = bool(i);
    CustomGame[LoadedGameNum].bNoEndGameVoting = bool(i);

    //---------------- sliders -------------------

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    EmptyServerSwitchTimeMins = i;
    default.EmptyServerSwitchTimeMins = i;

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    GameStartVoteDelay = i;
    default.GameStartVoteDelay = i;

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    VoteTimeLimit = i;
    default.VoteTimeLimit = i;

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    PopupWindowDelay = i;
    default.PopupWindowDelay = i;

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    RepeatLimit = i;
    default.RepeatLimit = i;

    i = InStr(Stuff, ",");
    sTemp = Left(Stuff, i);
    Stuff = Mid(Stuff, i+1);
    i = int(sTemp);
    TickRating = i;
    CheckTickRate(i, False);

    SaveConfig();

    if (MVCMD != None)
    {
        MVCMD.bKillBots = bAllowPlayerKillBots;
        MVCMD.bAddBot   = bAllowPlayerAddBot;
    }
}

function string GetAdminStuff()
{
    local string sTemp, s0, s1;

    s0 = "0,";
    s1 = "1,";

    if (bAutoOpen)               // wouldn't work any other way ?
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (bAllowMidGameVoting)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (bDebugMode)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (bAllowPlayerKillBots)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (bAllowPlayerAddBot)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (bMajorityPlayerVote)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (MVID.bUseSendToSpec)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (bEnableWelcomeMenu)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (bSortCustomGameList)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (bEmptyServerSwitch)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (bEmptyServerRandomMap)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    if (bNoEndGameVoting)
        sTemp = sTemp$s1;
    else
        sTemp = sTemp$s0;

    //---------------- sliders -------------------

    sTemp = sTemp$string(EmptyServerSwitchTimeMins)$","$string(GameStartVoteDelay)$","$string(VoteTimeLimit)$","$
            string(PopupWindowDelay)$","$string(RepeatLimit)$","$TickRating$",";

    return sTemp;
}

function CheckTickRate(int iTick, bool bGet)
{
    local string sTemp, sTick;
    local int i, j;

    sTemp = Level.ConsoleCommand("Get ipDrv.TcpNetDriver NetServerMaxTickRate");
    TickRating = int(sTemp);
    if (bGet)
        return;

    if (iTick != 0 && iTick != TickRating)
    {
        Level.ConsoleCommand("Set ipDrv.TcpNetDriver NetServerMaxTickRate "$iTick);
        Level.ConsoleCommand("Set ipDrv.TcpNetDriver LanServerMaxTickRate "$iTick);
        TickRating = iTick;

        sTemp = "";
	    if (DefaultSettings != "")
	    {
			i = InStr(Caps(DefaultSettings), "TICKRATE=");
			if (i >= 0)
			{
			    sTick = Mid(DefaultSettings, i);
			    j = InStr(sTick, ",");        //TickRate=80,whatever
			    if (j >= 0)
			       sTick = Left(sTick, j);    //TickRate=80
			    sTemp = Left(DefaultSettings, i)$Mid(DefaultSettings, i+j+1);
                if (Right(sTemp, 1) != ",")
                    sTemp = sTemp$",";
			}
		}
		DefaultSettings = sTemp$"TickRate="$iTick$",";
		SaveConfig();
	}
}

function OpenVoteWindow (PlayerPawn Sender, bool bSpecsAllowed, int iWelcomeFirst)
{
	local MapVoteWRI MVWRI, A;
	local int i, j, X;
	local int PlayerCount;
	local Pawn P;
	local int TeamID;
	local string sTemp;
	local bool bDone;

	if ( Sender.IsA('Spectator') &&  !CheckForAdmin(Sender) &&  !bSpecsAllowed )
	{
		return;
	}

	if ( bLevelSwitchPending )
	{
		return;
	}

    if (Level.Game == None)
        return;

	foreach AllActors(Class'MapVoteWRI',A)
	{
		if ( Sender == A.Owner )
		{
			return;
		}
	}

    GetServerInfo();             // ip and stuff

	MVWRI = Spawn(Class'MapVoteWRI',Sender,,Sender.Location);

	if ( MVWRI == None )
	{
		Log("#### -- PostLogin :: Fail:: Could not spawn WRI");
		return;
	}

    for (i=0; i<20; i++)
    {
         MVWRI.MyData[i] = "";
         if (i < 10)
             MVWRI.MyBool[i] = 0;
	}

    bDone = false;
    MVWRI.sCode = sCode;
    MVWRI.sTitle = "MapVoteX"$Version$"  -  Total maps loaded: "$LoadedMapNum;
    MVWRI.DefGame = CustomGame[DefaultCustomGameNum].GameName;
    MVWRI.bMapsOnly = bMapsOnly;
    MVWRI.bDebugMode = bDebugMode;
    MVWRI.PlayerNum = Level.Game.NumPlayers;
    MVWRI.bNoBots = true;
    if (DeathMatchPlus(Level.Game).NumBots > 0 && !Level.Game.bGameEnded && bAllowPlayerKillBots && !Sender.PlayerReplicationInfo.bWaitingPlayer)
        MVWRI.bNoBots = false;                      // flips Random Game button to Say !KillBots
    MVWRI.bSetupWindowDelayDone = false;

    if (MVCW == None)
        MVCW = Spawn(class'MapVoteClientWelcome');

    MVWRI.MyBool[1] = int(CheckForAdmin(Sender));

    MVWRI.MyBool[3] = 0;
    MVWRI.MyBool[4] = 0;
    if (bEnableWelcomeMenu)
    {
        MVWRI.MyBool[2] = int(bEnableWelcomeMenu);
        MVWRI.MyBool[3] = int(MVCW.GetItem(1));     // buse
        MVWRI.MyBool[4] = iWelcomeFirst;

        j = int(MVCW.GetItem(2));                   // bshowfirst
        if (j == 0 || MVWRI.MyBool[3] == 0)
            MVWRI.MyBool[4] = 0;                    // 2nd if default welcome
    }

    MVWRI.MyBool[5] = MapCount - GameCount;         // random maps
    MVWRI.MyBool[6] = int(bAllowMidGameVoting);
    MVWRI.MyBool[7] = int(bGameHasEnded);
    MVWRI.MyBool[8] = GameCount;                    // random games

    MVWRI.MyData[1] = sCode;
    MVWRI.MyData[2] = Version;

    if (CheckForAdmin(Sender))
    {
        sTemp = GetAdminStuff();
        MVWRI.MyData[3] = sTemp;
    }

    sTemp = ServerName;
    if (Len(sTemp) > 100)
        sTemp = Left(sTemp, 100);                      // ..:: PSYCHO | *NewNet* |
    MVWRI.MyData[4] = sTemp;
    MVWRI.MyData[5] = ServerPort;
    MVWRI.MyData[6] = ServerIP;
    MVWRI.MyData[8] = "";

    if (bEnableWelcomeMenu && MVWRI.MyBool[3] == 1)            // buse
    {
        MVWRI.MyData[7]  = MVCW.GetItem(3);     // stretch
        MVWRI.MyData[8]  = MVCW.GetItem(4);     // logoname
        MVWRI.MyData[9]  = MVCW.GetItem(5);     // soundname
    }

    MVWRI.MyData[12] = LoadedMapPrefix;

	PlayerCount=0;
	for (P = Level.PawnList; P != None; P = P.NextPawn)
	{
		if ((PlayerPawn(P) != None) && P.bIsPlayer)
		{
			if ( !P.IsA('Spectator') )
			{
 			    TeamID = 0;
 			    if (Level.Game.bTeamGame)
 			        TeamID = PlayerPawn(P).PlayerReplicationInfo.Team + 1;
			    if ( PlayerVoted[FindPlayerIndex(PlayerPawn(P).PlayerReplicationInfo.PlayerID, True)] > 0 )
			    {
				    MVWRI.PlayerName[PlayerCount++ ] = string(TeamID)$PlayerPawn(P).PlayerReplicationInfo.PlayerName;
				}
			}
		}
	}

    for (X=0; X<31; X++)
    {
 	    if (MapStatusText[X] != "")
	    {
		    MVWRI.MapVoteResults[X] = MapStatusText[X];
	    }
	}

	MVWRI.MapVoteResults[X]= "";

    if (bMapsOnly)
    {
        MVWRI.MapsOnlyCount = MapsOnlyCount;
        DebugLog("WRI Start - New Game Maps Count = "$MapsOnlyCount$" | Time: "$Level.TimeSeconds);
        return;
    }

    if (CheckForAdmin(Sender))
    {
        if (!bMapsOnly)
        {
            MVWRI.bAdminMesg = true;
            MVWRI.bMesgUpdate = false;

            if (MVAM == None)
                MVAM = Spawn(class'MapVoteAdminMessages');

            X=1;
	        for (i=0; i<50; i++)
	        {
			    if (MVAM.Message[i] == "")
			        break;
	            MVWRI.Mesglist[X] = MVAM.Message[i];
	            X++;
	        }
	    }
    }
    else
       MVWRI.bAdminMesg = false;
}

function bool AddBot(Pawn Sender)
{
    if (DeathMatchPlus(Level.Game).NumBots < (DeathMatchPlus(Level.Game).MaxPlayers - DeathMatchPlus(Level.Game).NumPlayers))
    {
        Level.Game.ForceAddBot();
        return true;
    }
    Sender.ClientMessage("Bots in game are already at maximum.");
    return false;
}

function bool RemoveBot(bool bAll)
{
    local Bot DumbBot;
    local Pawn P;

    for (P = Level.PawnList; P != None; P = P.NextPawn)
    {
        if (Bot(P) != None) // get Bot with least score
        {
           if (bAll)
           {
               Bot(P).Destroy();
               DeathMatchPlus(Level.Game).MinPlayers -= 1;
           }
           else
           if (DumbBot == None || P.PlayerReplicationInfo.Score < DumbBot.PlayerReplicationInfo.Score)
               DumbBot = Bot(P);
        }
    }

    if (DumbBot != None)
    {
        DeathMatchPlus(Level.Game).MinPlayers -= 1;
        DumbBot.Destroy();
        return true;
    }

    return false;
}

function int FindPlayerIndex (int PlayerID, bool bAddIfNotFound)
{
	local int i;

    for (i=0; i<32; i++)
	{
		if ( PlayerIDList[i] == PlayerID )
		{
			return i;
		}
	}

	if ( bAddIfNotFound )
	{
       for (i=0; i<32; i++)
	   {
		    if ( PlayerIDList[i] == -1 )
			{
				PlayerIDList[i] = PlayerID;
				return i;
			}
		}
	}
	return -1;
}

function CleanUpPlayerIDs()
{
	local Pawn aPawn;
	local int x;
	local bool bFound;

    for (x=0; x<32; x++)
    {
        if (PlayerIDList[x] > -1)
        {
           bFound = false;
           for ( aPawn=Level.PawnList; aPawn!=None; aPawn=aPawn.NextPawn )
           {
               if (aPawn.bIsPlayer && aPawn.IsA('PlayerPawn') && PlayerPawn(aPawn).PlayerReplicationInfo.PlayerID == PlayerIDList[x])
               {
                   bFound = true;
                   break;
               }
           }
		   if ( !bFound )
		   {
			   DebugLog("Removing Player that has left:" @ string(PlayerIDList[x]));
			   PlayerVoted[x] = 0;
			   PlayerRandM[x] = 0;
			   PlayerRandG[x] = 0;
			   PlayerIDList[x] = -1;
	       }
        }
    }
}

function ClearAllVoting()
{
    local int i;

    for (i=0; i<32; i++)
    {
         PlayerVoted[i] = 0;
         PlayerRandG[i] = 0;
         PlayerRandM[i] = 0;
         MapStatusText[i] = "";
    }
    RandomMapVote  = "";
    RandomGameVote = "";
}

function TallyVotes (bool bForceMapSwitch)
{
	local int   Index;
	local int   x, y;
	local int   topmap;
	local int   TieCount;
    local bool  bRandom, bRandGame;
	local string sTemp, sGame;
	local string CurrentMap;
	local float  VoteCount[TotalList];
	local int Ranking[32];
	local int PlayersThatVoted;
	local int PlayersLeftToVote;
	local int NumBots, NumPlayers;

	if ( bLevelSwitchPending )
	{
		return;
	}

    NumPlayers = Level.Game.NumPlayers;  // for local use.
    if (Level.NetMode == NM_StandAlone)
        NumPlayers = 1;
    DebugLog("Tally Start bForceMapSwitch = "$bForceMapSwitch$" | NumBots = "$NumBots$" | Numplayers = "$NumPlayers);

	PlayersThatVoted = 0;
	bGameVoted = false;
	bRandom = false;
    bRandGame = false;                      // random game

	for (x=0; x<32; x++)
	{
		if ( PlayerVoted[x] != 0 )
		{
			PlayersThatVoted++;
		    VoteCount[PlayerVoted[x]] += GetMapPriority(TotalMapList[PlayerVoted[x]]);

			DebugLog("PlayerVoted["$x$"]:" @ string(PlayerVoted[x]));
			DebugLog("VoteCount[PlayerVoted["$x$"]]:" @ string(VoteCount[PlayerVoted[x]]));
		}
	}

	if (!Level.Game.bGameEnded && !bMidGameVote && PlayersThatVoted > 0)
	{
		BroadcastMessage("Mid-Game Map Voting has been initiated !!!!");
		bMidGameVote = True;
		TimeLeft = VoteTimeLimit;
		ScoreBoardTime = 1;
		SetTimer(1.00, True);
		DebugLog("Tally Vote % = "$float(PlayersThatVoted) / float(NumPlayers) * 100$" | PlayersThatVoted = "$PlayersThatVoted$" | NumPlayers = "$NumPlayers$" | TimeLeft = "$TimeLeft);
	}

	Index=0;
	for (x = 1; x <= MapCount; x++)     // includes GameCount
	{
		if ( VoteCount[x] > 0 )
		{
			Ranking[Index++ ] = x;  // copy all map indexes to the ranking list if someone has voted it.
		}
	}

    // bubble sort ranking list by vote count
    for (x=0; x<index-1; x++)
    {
        for (y=x+1; y<index; y++)
        {
            if (VoteCount[Ranking[x]] < VoteCount[Ranking[y]])
            {
               topmap = Ranking[x];
               Ranking[x] = Ranking[y];
               Ranking[y] = topmap;
            }
        }
    }

	PlayersLeftToVote = NumPlayers - PlayersThatVoted;
	DebugLog("Tally NumPlayer: " $ string(NumPlayers) @ "PlayerVoted: " $ string(PlayersThatVoted) @ "PlayersToVote: " $ string(PlayersLeftToVote));

    if (bMajorityPlayerVote && (PlayersThatVoted > (NumPlayers / 2)))
    {
        PlayersLeftToVote = 0;
        DebugLog("Tally Majority votes forcing PlayersLeftToVote = 0");
    }

    if ((PlayersLeftToVote < VoteCount[Ranking[0]]) && (VoteCount[Ranking[1]] + PlayersLeftToVote * GetMapPriority(TotalMapList[Ranking[1]]) < VoteCount[Ranking[0]]) && (VoteCount[Ranking[2]] + PlayersLeftToVote * GetMapPriority(TotalMapList[Ranking[2]]) < VoteCount[Ranking[0]]) && (VoteCount[Ranking[3]] + PlayersLeftToVote * GetMapPriority(TotalMapList[Ranking[3]]) < VoteCount[Ranking[0]]))
	{
		bForceMapSwitch = True;
		DebugLog("Tally bForceMapSwitch = "$bForceMapSwitch);
	}

    //Update Status Page
    for (x=0; x<index; x++)
	{
		MapStatusText[X] = GetPureMapname(TotalMapList[Ranking[x]]);
		if (MapStatusText[X] == RandomMapVote)
		    MapStatusText[X] = "[RR]"$MapStatusText[X];
		if (MapStatusText[X] == RandomGameVote)
		    MapStatusText[X] = "[RG]"$MapStatusText[X];
		MapStatusText[X] = MapStatusText[X]$","$Left(string(VoteCount[Ranking[x]]),3);
		DebugLog("Tally MapStatusText["$X$"] = "$MapStatusText[X]);
	}
	MapStatusText[Index]= "";

    UpdateOpenWRI(False);

	//Check for a tie
	if ( (VoteCount[Ranking[0]] == VoteCount[Ranking[1]]) && (VoteCount[Ranking[0]] != 0) )
	{
	    TieCount = 1;
        for (x=1; x<index; x++)
        {
            if (VoteCount[Ranking[0]] == VoteCount[Ranking[x]])
               TieCount++;
        }

	    topmap = Ranking[Rand(TieCount)];
	    // Don't allow same map to be choosen
		CurrentMap = RemovePrefix(GetURLMap());
        DebugLog("Tally Topmap = "$Topmap$" | CurrentMap = "$CurrentMap);
		if ( InStr(CurrentMap,".") > -1 )
		{
			CurrentMap = Left(CurrentMap,InStr(CurrentMap,"."));
		}

        x = 0;
        while (TotalMapList[topmap] ~= CurrentMap)
        {
              topmap = Ranking[Rand(TieCount)];
              x++;
              if (x > 20)
                 break;
        }
	}
	else
	{
		topmap = Ranking[0];
	}

    DebugLog("Tally bMapsOnly = "$bMapsOnly$" | bForceMapSwitch = "$bForceMapSwitch$" | bMidGameVote = "$bMidGameVote$" | TimeLeft = "$TimeLeft$" | PlayersThatVoted = "$PlayersThatVoted$" | bMidGameTimeout = "$bMidGameTimeout$" | Topmap = "$TotalMapList[topmap]);

	if ( bForceMapSwitch )
	{
		if ( (PlayersThatVoted == 0) || (NumPlayers == 0) )
		{
			   if (MVGRI.bCoopGame)
			   {
				   ClearAllVoting();
				   CloseAllVoteWindows();
			       sTemp = CustomGame[DefaultCustomGameNum].DefaultMap;
			       BroadcastMessage("No votes received, switching to default map: "$sTemp, True);
                   ServerTravelString = SetupGameMap(sTemp); // returns MapName $ ".unr?Game=" $ GameClass;
		           ServerTravelTime = Level.TimeSeconds + 2;
		           bLevelSwitchPending = True;
		           SetTimer(1.0, true);
		           return;
			   }
			   else
			   {
			       Topmap = GetRandomMap();
			       bRandom = true;

			       if (Level.Game.bGameEnded && bNoEndGameVoting && !bMapsOnly)
			       {
			           if (MapCount-GameCount == 1)
			               BroadcastMessage("Reloading map: "$TotalMapList[topmap], True);
			           else
			               BroadcastMessage("Switching to the Random map: "$TotalMapList[topmap], True);
			           goto Tallyz;
			       }
		           BroadcastMessage("No votes received, choosing a random map...", True);
		           DebugLog("Tally No votes received - Random Map = "$topmap$" | bmapsOnly = "$bMapsOnly);
		       }
		}
	}

	if ( bForceMapSwitch || (NumPlayers == PlayersThatVoted) )
	{
		if ( TotalMapList[topmap] == "" )
		{
		    DebugLog("Tally Topmap = NULL - Return");
			return;
		}

        if ( bMidGameTimeout )
        {
			// clear everything and close windows.
			bMidGameTimeout = false;
			bForceMapSwitch = false;
			bMidGameVote = false;
			ClearAllVoting();
			CloseAllVoteWindows();

		    if (bMapsOnly) // a new game ?
		    {
                topmap = GetRandomMap();
                if (TotalMapList[topmap] == "")
                    topmap = GetRandomMap();
                bRandom = true;
                BroadcastMessage(GetPureMapname(sTemp)$" has randomly won the vote.", True);
                DebugLog("Tally - MapsOnly - Random Selection - Map["$topmap$"] = "$TotalMapList[topmap]);
		    }
		    else
		    {
		        if ( NumPlayers == PlayersThatVoted )
 		            BroadcastMessage("Tie:  Midgame votes have been cancelled.", True);
		        else
		            BroadcastMessage("Timeout: Midgame votes have been cancelled.", True);
		        DebugLog("Tally - Tie or Timeout:  Midgame votes have been cancelled.");
		        return;
		    }
        }

        sTemp = TotalMapList[topmap];
        DebugLog("Tally TotalMapList[topmap] ["$topmap$"] = "$sTemp);

		if (Left(sTemp, 4) == "[GG]")
		{
		    bGameVoted = true;
		    sTemp = Mid(sTemp, 4);
		    if (sTemp ~= RandomGameVote)
		        bRandGame = true;
        }

		if (Left(sTemp, 4) == "[RG]")
		{
		    bGameVoted = true;
		    sTemp = Mid(sTemp, 4);
        }

		if ( (VoteCount[Ranking[0]] == VoteCount[Ranking[1]]) && (VoteCount[Ranking[0]] != 0) )
		{
			if (!Level.Game.bGameEnded && !bMapsOnly)
			{
		       DebugLog("Tally - TimeLeft = "$TimeLeft$" | No vote has won - Exit");
		       return;
			}
			BroadcastMessage("Tie: "$GetPureMapname(sTemp) $ " has won.", True);
			DebugLog("Tally Vote drew a tie, but the Map: "$sTemp$" has won.");
		}
		else
		{
		    sTemp = GetPureMapname(sTemp);
		    if (sTemp ~= RandomMapVote)
		        BroadcastMessage("A Random Map has won the vote.", True);
		    else
		    if (sTemp ~= RandomGameVote)
		        BroadcastMessage("A Random Game has won the vote.", True);
		    else
		    {
		        BroadcastMessage(sTemp$" has won the vote.", True);
		        if (!bRandom)
		        {
  		            if (bGameVoted)
		                MVTG.SetGameStuff(sTemp);
		            else
		                MVTM.SetMapStuff(sTemp);
		        }
		    }
		}

        if (!bGameVoted || !bQuickGameVote)
		    CloseAllVoteWindows();
        //**** Tally - The Winning Vote was: One Map DeathMatch | bGameVoted = True
        DebugLog("Tally - The Winning Vote was: "$sTemp$" | bGameVoted = "$bGameVoted);

        if (bGameVoted)
        {
            CheckForGameName(sTemp);
            if (LoadedGameNum >= 0)
            {
                sGame = CustomGame[LoadedGameNum].GameClass;

                if (bRandGame)
                {
                    sTemp = TotalMapList[topmap];
					ServerTravelString = SetupGameMap(sTemp);
					ServerTravelString = ServerTravelString$GetInlineSettings(0, sTemp);
					CloseAllVoteWindows();
					ServerTravelTime = Level.TimeSeconds + 2;
		            bLevelSwitchPending = True;
		            SetTimer(1.0, true);
		            DebugLog("Admin Forced Game TravelString: "$ServerTravelString);
		            return;
                }

                VotedGameClass = "?"$sGame;
                if (bQuickGameVote)
                {
                    DebugLog("Tally UpdateOpenWRI(True)");
                    UpdateOpenWRI(True);
                    BroadcastMessage("Please wait for "$sTemp$" Maps to load.", False);
                }
                else
                {
                    BroadcastMessage(" ", True);
                    BroadcastMessage("Please wait for "$sTemp$" Maps to load.", False);
                }

				ClearAllVoting();
                bMapsOnly = true;
                bMapsLoaded = false;
                ClearGRIMaps();
                SetupGameMap(TotalMapList[topmap]); //[GG]Assault - //AS-Frigate.unr?Game=Botpack.Assault
                DebugLog("Tally - VotedGameClass = "$VotedGameClass$" | New Game Maps Loading.....");
				DeathMatchPlus(Level.Game).bDontRestart = True;
				CheckMaxTeams(LoadedGameNum);
			    TimeLeft = 0;
				bLevelSwitchPending = false;
				ScoreBoardTime = 0;          //PopupWindowDelay will set scoreboardtime again
				if (!bQuickGameVote)
				{
				    CloseAllVoteWindows();
			        SetTimer(1.00,True);
			    }
                GotoState('Loading');
				return;
            }
        }
Tallyz:

		if (VotedGameClass != "")
		{
		    ServerTravelString = sTemp$VotedGameClass;
		    VotedGameClass = "";
		}
		else
		{
		    ServerTravelString = SetupGameMap(TotalMapList[topmap]);          //CTF-Face.unr?Game=Botpack.CTFGame
		}

		bLevelSwitchPending = True;
        ServerTravelString = SetupGameMap(TotalMapList[topmap]);              //CTF-Face.unr?Game=Botpack.CTFGame
        if (bMapsOnly)
            ServerTravelString = ServerTravelString$GetInlineSettings(LoadedGameNum, "");
		ServerTravelTime = Level.TimeSeconds + 4;
		if (RepeatLimit > 0)
            MVRP.AddMap(LoadedGameNum, TotalMapList[topmap]);

		DebugLog("Tally - ServerTravelString = "$ServerTravelString$" | bMapsOnly = "$bMapsOnly);

		if ( !Level.Game.bGameEnded )
		{
			SetTimer(1.00,True);
		}

		if ( PlayersThatVoted == 0 || NumPlayers == 0 )
		{
			return;
		}
	}
	DebugLog("Tally - Exit");
}

function CheckMaxTeams(int gNum)
{
    local string sClass, sSetts;
    local int i, j;

    sClass = CustomGame[gNum].GameClass;
    sSetts = CustomGame[gNum].Settings;

    i = InStr(Caps(sSetts), "MAXTEAMS=");
    if (i == -1)
        return;

    j = int(Mid(sSetts, i + 9));

    i = InStr(Caps(sClass), "ASSAULT");
    if (i >= 0)
    {
        class'Botpack.Assault'.default.MaxTeams = j;
        return;
    }
    i = InStr(Caps(sClass), "DOMINATION");
    if (i >= 0)
    {
        class'Botpack.Domination'.default.MaxTeams = j;
        class'Botpack.TeamGamePlus'.default.MaxTeams = j;
        return;
    }
    i = InStr(Caps(sClass), "CTFGAME");
    if (i >= 0)
    {
        class'Botpack.CTFGame'.default.MaxTeams = j;
        class'Botpack.CTFGame'.static.StaticSaveConfig();
        return;
    }
    i = InStr(Caps(sClass), "TEAMGAMEPLUS");
    if (i >= 0)
    {
        class'Botpack.TeamGamePlus'.default.MaxTeams = j;
        //class'Botpack.TeamGamePlus'.static.StaticSaveConfig();
    }
}

function UpdateOpenWRI(bool bTitle)
{
	local MapVoteWRI MVWRI;
	local Pawn P;
	local int X, Z, Id;
	local int TeamID;

	foreach AllActors(Class'MapVoteWRI',MVWRI)
	{
		Z = 0;
		for (P = Level.PawnList; P != None; P = P.NextPawn)
		{
			if ((PlayerPawn(P) != None) && P.bIsPlayer)
			{
				if ( !P.IsA('Spectator') )
				{
 	 			    TeamID = 0;
 	 			    if (Level.Game.bTeamGame)
 	 			        TeamID = PlayerPawn(P).PlayerReplicationInfo.Team + 1;
 	 			    Id = FindPlayerIndex(PlayerPawn(P).PlayerReplicationInfo.PlayerID, True);
				    if ( PlayerVoted[Id] > 0 )
				    {
					    MVWRI.PlayerName[Z++] = string(TeamID)$PlayerPawn(P).PlayerReplicationInfo.PlayerName;
					    if (bTitle)
					    {
					        MVWRI.sTitle = "Please wait for '"$CustomGame[LoadedGameNum].GameName$"' Maps to load.";
					        MVWRI.UpdateMapVoteResults(MVWRI.sTitle, X);
					    }
					}
				}
			}
		}

		MVWRI.UpdateMapVoteResults("Clear", X);
		for (X=0; X<31; X++)
		{
		    if (MapStatusText[X] != "")
		    {
 		        MVWRI.UpdateMapVoteResults(MapStatusText[X], X);  // [RR]CTF-Niven,1.0
 		    }
		}
		MVWRI.UpdateMapVoteResults("", X);
	}
}

function float GetMapPriority(string MapName)
{
	if (MapName == "")
		return 0.00;
	else
	    return 1.00;
}

function string GetPureMapname (string MapName)
{
    local string sMap;
    local int j;

	if ( Left(MapName,4) ~= "[XX]" )
	{
		MapName=Mid(MapName,4);
	}
	else
	if ( Left(MapName,4) == "[RR]" )                       // only when 1st player votes random
	{
	   sMap = Mid(MapName, 4);
       j = GetRandomMap();
       MapName = TotalMapList[j];
       if (Mapname ~= sMap)
       {
           j = GetRandomMap();                             // try once more
           MapName = TotalMapList[j];
       }
       if (MapName == "")
           MapName = CustomGame[LoadedGameNum].DefaultMap;
	}
	else
	if ( Left(MapName,4) ~= "[RG]" )                       // [RG]Random
	{
       MapName = GetRandomGame();
       if (MapName == "")
           MapName = CustomGame[DefaultCustomGameNum].GameName;
	}
	else
	if ( Left(MapName,4) ~= "[GG]" )
	{
		MapName=Mid(MapName,4);
	}
	else
	if ( Left(Right(MapName,2),1) == ":" )
	{
		MapName=Left(MapName,Len(MapName) - 2);
	}
	return MapName;
}

function CloseAllVoteWindows ()
{
	local MapVoteWRI MVWRI;

	foreach AllActors(Class'MapVoteWRI',MVWRI) // chances are window is already closed
	{
		MVWRI.CloseWindow();
		MVWRI.Destroy();
	}
}

function HandleDefaultSwitch(bool bAdm)
{
    local string sMap;
    local int i, j;

    i = DefaultCustomGameNum;
    if (i < 0 || i > MaxNumGames-1)
        i = 0;

    if (bEmptyServerSwitch || bAdm)
    {
        if (!bAdm && bEmptyServerRandomMap)
        {
	        j = GetRandomMap();
	        sMap = TotalMapList[j];
		    ServerTravelString = SetupGameMap(sMap);
		}
        else
        if (MVGameName != CustomGame[i].GameName)
        {
            if (CustomGame[i].bUse == False || CustomGame[i].DefaultMap == "" || CustomGame[i].GameClass == "" || CustomGame[i].GameName == "")
            {
                Log("Empty Server Switch cannot switch due to error in Default CustomGame["$i$"]", 'MapVoteX');
                return;
            }
            ServerTravelString = SetupGameMap("[GG]"$CustomGame[i].GameName);
            sMap = CustomGame[i].DefaultMap;
        }
        else
        if (bEmptyServerSwitchMap)
        {
            ServerTravelString = SetupGameMap(CustomGame[i].DefaultMap);
            sMap = CustomGame[i].DefaultMap;
        }
        else
            return;

        if (LoadedMapName ~= CustomGame[i].DefaultMap || ServerTravelString == "")
            if (!bEmptyServerRandomMap)
                return;

		ServerTravelTime = Level.TimeSeconds + 2;
		bLevelSwitchPending = True;
		SetTimer(1.0, true);
        Log("Empty Server Switch at: "$GetAbsoluteTime()$" to: "$CustomGame[i].GameName$" and Map: "$sMap, 'MapVoteX');
    }
}

function string SetupGameMap (string MapName)
{
    local string sTemp;
	local string GameType;
	local string GameName;
	local bool bGame;
	local int i, j, iNum;


    if (Left(Mapname, 4) == "[GG]")
        bGame = true;

    if (Left(Mapname, 4) == "[RG]")   // never
        bGame = true;

    MapName=GetPureMapname(MapName);
	ResetAssault();
    iNum = -1;

    GameType = string(Level.Game.Class);
    GameName = Level.Game.GameName;

	if ( MapName == "" )
	{
		return "?Restart";
	}

    if (bGame)
	{
	    sTemp = MapName; // [GG] removed
	    MapName = "";
	    for (i=0; i<MaxNumGames; i++)
	    {
	        if (CustomGame[i].GameName ~= sTemp)
	        {
	           iNum = i;
	           GameType = CustomGame[i].GameClass;
	           MapName  = CustomGame[i].DefaultMap;
               j = InStr(Caps(MapName),".UNR");
               if ( j != -1 )
		          MapName = Left(MapName, j);
	           GameName  = sTemp;
	           LoadedMapPrefix = GetMapPrefix(MapName);
	           break;
	        }
	    }
        if (iNum >= 0)
        {
            SetMapPath(CustomGame[iNum].MapFolder);
            SavePackageActors(iNum);
        }
	}
	else
	{
	     for (i=0; i<MaxNumGames; i++)
	     {
	          if (CustomGame[i].GameName ~= MVGameName)
	          {
	              iNum = i;
	              GameType = CustomGame[i].GameClass;
	              GameName = CustomGame[i].GameName;
	              break;
	          }
	     }
	}

    SaveGameInfo(GameName);
	SaveConfig();
	TravelMap = MapName;
	return MapName $ ".unr?Game=" $ GameType;
}

state Loading
{
Begin:

    MapCount = 0;
    MapTime = Level.TimeSeconds;

    if ( !bLevelSwitchPending && !bRestarting )
    {
		for ( m=0; m<TotalList; m++ )
		{
			TotalMapList[m] = "";
		}

		for (m=0; m<MaxNumGames; m++)
		{
			if ( CustomGame[m].bUse && (CustomGame[m].GameClass != "") && (CustomGame[m].GameName ~= MVGameName))
			{
				DebugLog("State Loading: Loading maps for: "$CustomGame[m].GameName$" Path: "$CustomGame[m].MapFolder);
				CustomGameClass=Class<GameInfo>(DynamicLoadObject(CustomGame[m].GameClass,Class'Class'));
				if ( CustomGameClass == None )
				{
					Log("MapVote: Error. CustomGame[" $ string(m) $ "]: Failed loading GameClass:" @ CustomGame[m].GameClass);
					break;
				}
				else
				{
					if (CustomGame[m].DefaultMap != "")
					{
					    NewPrefix = FixPrefix(GetMapPrefix(CustomGame[m].DefaultMap));
					}
					else
					{
						NewPrefix = FixPrefix(CustomGameClass.default.MapPrefix);
					}

					Level.Game.GameName = CustomGame[m].GameName;

					if ((bMapsOnly || MVML == None || VotedGameClass != "") && !bLevelSwitchPending)
					{
 					    if (MVML != None && VotedGameClass != "")
					        MVML.Destroy();
					    bMapsLoaded = false;
			            MVML = Spawn(class'MapVoteMapLoader');
                        MVML.MyMut = self;
                        MVML.MaxNumMaps = MaxNumMaps;
                        MVML.DefClas  = CustomGame[m].GameClass;
                        MVML.DefPath  = CustomGame[m].MapFolder;
			            MVML.Prefix = NewPrefix;
			            DebugLog("State Loading Maps | NewPrefix = "$NewPrefix$" | GameName = "$Level.Game.GameName$" | GameNum = "$m$chr(10)$chr(13));
 		                MVML.StartLoadingMaps(NewPrefix, m, CustomGame[m].GameName, CustomGame[m].MapFolder);
					}

					Sleep(0.00);
					break;
				}
				Sleep(0.00);
			}
		}
    }
}

function SortMapsAndGames()    // called from maploading class
{
    local int i, j;
    j = ScoreBoardTime;
    LoadGameList();
    i = MapCount-GameCount;
    LoadedMapNum = i;          // test for single map later
    CheckServername();
    GetServerInfo();
    LoadGameReplication();
    if (bMapsOnly)
    {
        if (bQuickGameVote)
        {
            if (LoadedMapNum < 100)
                ScoreBoardTime = 5;
            else
            if (LoadedMapNum < 200)
                ScoreBoardTime = 10;
            else
            if (LoadedMapNum < 500)
                ScoreBoardTime = 15;
            else
                ScoreBoardTime = 20;
            SetTimer(1.0, True);
        }
        else
            ScoreBoardTime = FMax(PopupWindowDelay, 15);
    }
    Log("Map Loaded: "$LoadedMapName$" | Total Maps Loaded: "$MVGRI.LoadedMapNum$" | Load Time: "$Level.TimeSeconds - MapTime, 'MapVoteX');
}

function LoadGameList()
{
    local string sGame, sMap;
    local int i, j, k;

    for (i=0; i<MaxNumGames; i++)
    {
        GameList[i] = "";
        GameMaps[i] = "";
    }

    k = 0;
    GameCount = 0;
	for (i=0; i<MaxNumGames; i++)
	{
	    if (CustomGame[i].bUse && CustomGame[i].GameName != "")
	    {
		    GameList[k] = "[GG]"$CustomGame[i].GameName;
		    sMap = CustomGame[i].DefaultMap;
            j = InStr(Caps(sMap),".UNR");
            if ( j != -1 )
                sMap = Left(sMap, j);
		    if (sMap == "")
		        sMap = "None";
		    GameMaps[k] = sMap;
		    GameCount++;
		    k++;
		}
	}

   if (bSortCustomGameList)
   {
      // bubble sort the game list
      for(i=0; i<=GameCount-1; i++)
      {
         for(j=i+1; j<=GameCount; j++)
         {
            if(Caps(GameList[i]) > Caps(GameList[j]))
            {
               sGame       = GameList[i];
               GameList[i] = GameList[j];
               GameList[j] = sGame;

               sMap        = GameMaps[i];
               GameMaps[i] = GameMaps[j];
               GameMaps[j] = sMap;
            }
         }
      }
   }

   if (GameList[0] == "" && GameList[1] != "")  // game sorting moves all up 1
   {
       for (i=0; i<MaxNumGames; i++)
       {
            GameList[i] = GameList[i+1];
            GameMaps[i] = GameMaps[i+1];
       }
   }

   j = 0;
   for (i=0; i<TotalList; i++)
   {                      // [GG]GameName's are added immediately after maps
       if (i > 0 && TotalMapList[i] == "" && GameList[j] != "" && j < MaxNumGames)
       {
           TotalMapList[i] = GameList[j]; // use at tallyvote only
           MapCount++;
           j++;
       }
   }
}

function UpdateGRI()
{
    local string sTemp;
    local int i;

	for (i=0; i<MaxNumGames; i++)
	{
	     sTemp = CustomGame[i].GameName;
	     if (sTemp == "")
	         sTemp = "None";
	     if (CustomGame[i].bUse && sTemp != "None")
	         sTemp = "*"$sTemp;
         MVGRI.GameNames[i] = sTemp;
    }
}

function ClearGRIMaps()
{
    local int i;

    for (i=0; i<MaxNumMaps; i++)
    {
        if (i < 256)
            MVGRI.ListedMaps1[i] = "";
        else
        if (i < 512)
            MVGRI.ListedMaps2[i-256] = "";
        else
        if (i < 768)
            MVGRI.ListedMaps3[i-512] = "";
        else
        if (i < 1024)
            MVGRI.ListedMaps4[i-768] = "";
    }
}

function LoadGameReplication()
{
    local string sTemp, sUrl;
    local int i, j, X;

    if (MVGRI == None)
        MVGRI = Spawn(class'MapVoteGRI');

    if ( bMapsOnly )
    {
        MVGRI.MapPref = LoadedMapPrefix;
        MVGRI.sCode = sCode;
        MVGRI.GameName = MVGameName;
        MVGRI.MaxNumMaps = MaxNumMaps;
        MVGRI.bMapsOnly = True;
    }
    else
    {
        MVGRI.sCode = sCode;
        MVGRI.MaxNumMaps = MaxNumMaps;
        MVGRI.bDoFavor = bAddToFavorites;
        MVGRI.bMapsOnly = False;

        j = 1;
        // do the game maps 1st
        for (i=0; i<MaxNumGames; i++)
        {
            if (GameList[i] == "")
                break;
            MVGRI.GameStuff1[j] = GameList[i];
            MVGRI.GameStuff2[j] = GameMaps[i];
            j++;
        }

        MVGRI.GameStuff2[202] = MVGameName;
        MVGRI.GameStuff2[203] = LoadedMapPrefix;
        MVGRI.GameStuff2[204] = string(GameCount);
    }

    MVGRI.bDoubleClickVoting = bDoubleClickVoting;

    // do the map names
    j = Len(LoadedMapPrefix);
    X = 0;

    MVGRI.RestartMap = LoadedMapName;
    MVGRI.LoadedMapNum = LoadedMapNum;
    MVGRI.MaxNumMaps = MaxNumMaps;

	for (i=0; i<MaxNumMaps+1; i++)
	{
		if (TotalMapList[i] != "")
		{
		   sTemp = TotalMapList[i];

		   if (Left(sTemp, 4) != "[GG]")
		   {
		      sTemp = Mid(sTemp, j);                                   // removes prefix from map for replication

		      if (!bMapsOnly)
		      {
		         if (RepeatLimit > 0)
		         {
		             if (MVRP.CheckMap(LoadedGameNum, TotalMapList[i]))
		             {
		                 sTemp = "[XX]"$sTemp;
		                 MVGRI.RestartMap = "";
		             }
		         }
		      }

		      if (X < 256)
		      {
	              MVGRI.ListedMaps1[X] = sTemp;
		      }
		      else
		      if (X < 512)
		      {
	              MVGRI.ListedMaps2[X-256] = sTemp;
		      }
		      else
		      if (X < 768)
		      {
	              MVGRI.ListedMaps3[X-512] = sTemp;
		      }
		      else
		      if (X < 1024)
		      {
	              MVGRI.ListedMaps4[X-768] = sTemp;
		      }
		      X++;
		   }
		}
		if ((i > (MapCount + 2)) && TotalMapList[i] == "")
		   break;
	}

    UpdateGRI();

    if (bMapsOnly)
    {
        MapsOnlyCount = X;
        return;
    }

    if (bEnableWelcomeMenu)
    {
        if (MVWL == None)
        {
            MVWL = Spawn(class'MapVoteWelcomeLinks');
            MVWL.MyMut = self;
        }

        for (i=0; i<25; i++)
        {
             sTemp = MVWL.GetLinkInfo(i, 0);
             sUrl  = MVWL.GetLinkInfo(i, 1);
             MVGRI.DropSetup(i, sTemp, sUrl);
        }
    }
}

function string FixPrefix (string Prefix)
{
	if (Right(Prefix, 1) == "-" || Right(Prefix, 1) == "+")
		return Prefix;
	else
		return Prefix$"";
}

function string RemovePrefix (string MapName)
{
	if (Left(MapName,7) ~= "CTF-BT-")   // CTF-BT+
	    return Mid(MapName, 7);
	if (Left(MapName,7) ~= "CTF-BT+")   // CTF-BT+
	    return Mid(MapName, 7);
	if (Left(MapName,3) ~= "BT-")       // CTF-BT+
	    return Mid(MapName, 3);
	if (Left(MapName,3) ~= "BT+")       // CTF-BT+
	    return Mid(MapName, 3);
	return Mid(MapName,InStr(MapName, "-") + 1);
}

function string GetMapPrefix(string MapName)
{
    local int j;

	if (Left(MapName,7) ~= "CTF-BT-")
		return "CTF-BT-";
	if (Left(MapName,7) ~= "CTF-BT+")
		return "CTF-BT+";
	if (Left(MapName,3) ~= "BT-")
		return "BT-";
	if (Left(MapName,3) ~= "BT+")
		return "BT+";

    j = InStr(MapName, "-");
    if (j == -1)
        return "";
    else
        return Left(MapName, j+1);
}

function SetMapPath(string MapPath)
{
    local string sTemp, sPath, OldPath, sFinal;
    local string Paths[MaxNumGames];
    local int i, j, k, x;

    for (i=0; i<MaxNumGames; i++)
         Paths[i] = "";

    if (Caps(MapPath) == "MAPS" || Caps(MapPath) == "MAPS/" || Caps(MapPath) == "../MAPS")
        MapPath = "";

    x = 5;
    if (MapPath != "")
    {
        if (Right(MapPath, 1) != ",")
            MapPath = MapPath$",";
        for (i=0; i<x; i++)
        {
             j = InStr(MapPath, ",");
             if (j == -1)
                 break;
             Paths[i] = Left(MapPath, j);
             MapPath = Mid(MapPath, j+1);
        }
    }

    sPath = Level.ConsoleCommand("Get Core.System Paths");
    OldPath = sPath;

    i = InStr(Caps(sPath), "../SYSTEM/*.U");
    if (i > 5)
    {
	    Log("System.Core Failure! Paths=../System/*.u is not first. Cannot write to file!", 'MapVoteX');
	    Log("Current Order: "$sPath, 'MapVoteX');
        if (MVAM == None)
            MVAM = Spawn(class'MapVoteAdminMessages');
        MVAM.SaveAdminMessage("MVX", "System.Core Failure! Paths=../System/*.u is not first. Cannot write to file!");
    }

    j = InStr(sPath, ")");
    if (j >= 0)
       sPath = Left(sPath, j);
    sPath = sPath$",";
    sFinal = "";
    for (i=0; i<MaxNumGames; i++)
    {
        j = InStr(sPath, ",");
        if (j < 0)
            break;
        sTemp = Left(sPath, j);
        sPath = Mid(sPath, j+1);
        k = InStr(Caps(sTemp), ".UNR");
        if (k < 0)
        {
           sFinal = sFinal$sTemp$",";
        }
        else
        {
           k = InStr(Caps(sTemp), "MAPS/*.UNR");
           if (k > 0)
              sFinal = sFinal$sTemp$",";
        }
    }

    for (i=0; i<x; i++)
    {
        if (Paths[i] != "")
        {
            sTemp = "\"../"$Paths[i]$"/*.unr\",";
            j = InStr(Caps(sFinal), Caps(sTemp));
            if (j == -1)
                sFinal = sFinal$sTemp;
        }
    }

    if (Right(sFinal, 1) == ",")
        sFinal = Left(sFinal, Len(sFinal)-1);
    sFinal = sFinal$")";

    if (sFinal ~= OldPath)
        return;

    Level.ConsoleCommand("set Core.System Paths "$sFinal);

}

function string GetAllPaths()
{
     local string sTemp, sPath;
     local int i, j;

     sPath = "";

     for (i=0; i<MaxNumGames; i++)
     {
         if (CustomGame[i].bUse == True)
         {
            sTemp = CustomGame[i].MapFolder;
            if (sTemp != "")
            {
                if (Caps(sTemp) != "MAPS" && Caps(sTemp) != "MAPS/")
                {
                    j = InStr(Caps(sPath), Caps(sTemp));
                    if (j == -1)
                        sPath = sPath$sTemp$",";
                }
            }
         }
     }
     return sPath;
}

function DebugLog (string Msg)
{
	if ( bDebugMode )
	{
		Log("**** " $Msg);
	}
}

function int ResetAssault()
{
	local int CurrentDefender;

	if ( Level.Game.IsA('Assault') )
	{
		if ( GetWinningTeam() != -1 )
		{
			Log("MapVoteX: Team " $ string(GetWinningTeam()) $ " won the round.");
			if ( GetWinningTeam() == 1 )
			{
				CurrentDefender=0;
			}
			else
			{
				CurrentDefender=1;
			}
			Log("MapVoteX: Setting Team " $ string(CurrentDefender) $ " as the defending team for the next map.");
		}
		else
		{
			CurrentDefender=0;
		}
		Assault(Level.Game).ResetGame();
		Assault(Level.Game).bDefenseSet=False;
		Assault(Level.Game).NumDefenses=0;
		Assault(Level.Game).SavedTime=0.00;
		Assault(Level.Game).GameCode="";
		Assault(Level.Game).Part=1;
		Assault(Level.Game).bTiePartOne=False;
		Assault(Level.Game).CurrentDefender=CurrentDefender;
		Assault(Level.Game).SaveConfig();
	}
	return 0;
}

function int GetWinningTeam ()
{
	if ( Assault(Level.Game).bDefenseSet )
	{
		if ( Assault(Level.Game).bAssaultWon )
		{
			return Assault(Level.Game).Attacker.TeamIndex;
		}
		else
		{
			if (  !Assault(Level.Game).bTiePartOne )
			{
				return Assault(Level.Game).Defender.TeamIndex;
			}
			else
			{
				return -1;
			}
		}
	}
	else
	{
		return -1;
	}
}

function int GetMapIndex (string MapName)
{
	local int i;
	local int MapIndex;

    MapName = GetPureMapname(MapName);

	for (i=1; i<=MapCount; i++)      // includes GameCount
	{
	   if ( GetPureMapname(TotalMapList[i]) ~= MapName ) // includes game names
	   {
		  MapIndex=i;
		  break;
	   }
	}

	return MapIndex;
}

function int GetRandomMap()
{
	local int i, R, x;
	local string sTemp;

    x = MapCount - GameCount;
    if (x == 1)
        return 1;

    for (i=0; i<100; i++)
    {
        R = Rand(x);
        sTemp = GetPureMapname(TotalMapList[R]);
	    if ( R > 0 && (LoadedMapName != sTemp) )
	       break;
	}

	if (R < 1)
	    R = 1;

	return R;
}

function string GetRandomGame()
{
	local string sTemp;
	local int i, R, x;

    x = GameCount;

    for (i=0; i<110; i++)
    {
        R = Rand(x);
        sTemp = Mid(GameList[R], 4);   // remove [GG]
        if (sTemp != Level.Game.GameName)
        {
            if (!CheckNoRandomGame(sTemp))
                 break;
        }
	}

    if (i >= 109)
        return Level.Game.GameName;

    for (i=0; i<MaxNumGames; i++)
    {
        if (sTemp == CustomGame[i].GameName)
            break;
    }

	return sTemp;
}

function bool CheckNoRandomGame(string sTemp)
{
    local int i;

    for (i=0; i<MaxNumGames; i++)
    {
        if (sTemp == CustomGame[i].GameName)
        {
            if (CustomGame[i].bNoRandomGame)
                return True;
            else
                return False;
        }
    }
	return False;
}

function bool InitializeMvx()
{
    local string sMap, sSet, sTemp;
    local int i;

    i = DefaultCustomGameNum;
    if (i < 0 || i > MaxNumGames-1)
        i = 0;

    SaveConfig();

    sMap = CustomGame[i].DefaultMap;
    if (sMap != "")
    {
        SpawnAdvancedClasses();
        bRestarting = true;
        if (MVML == None)
            MVML = Spawn(class'MapVoteMapLoader');

        SetMapPath(CustomGame[i].MapFolder);
        SaveGameInfo(CustomGame[i].GameName);
        sTemp = sMap$"?Game="$CustomGame[i].GameClass;  // CTF-Face?Game=Botpack.CTFGame
        sSet = GetInlineSettings(i, "");
        if (sSet != "")
            sTemp = sTemp$sSet;
        Log("Server Startup - Restarting Default Game: "$sTemp, 'MapVoteX');
        ServerTravelString = sTemp;
        sTemp = CustomGame[i].ServerName;
        if (ServerNameTitle != "")
            sTemp = ServerNameTitle$sTemp;
        if (bUseServerNames && sTemp != "")
            SetServername(sTemp);
        SavePackageActors(i);
        Level.ServerTravel(ServerTravelString,False);
        return false;
    }

    Log("#### MapVoteX: Cannot restart the Default Game: "$CustomGame[i].GameName$" as it's DefaultMap: "$sMap$" cannot be found!");
    HandleFailedMap("Default Map");
    return true;
}

function bool ExecuteMutatorsAndSettings()
{
	local string Mutators;
	local string Settings;
	local string sGame, sTemp, sSet, sAly;
	local int i, j, k, x, pos;
	local bool bFound;

    if (MVGA == None)
        MVGA = Spawn(class'MapVoteGameAliases');

    j = InStr(Caps(LoadedMapName), "DM-TUTORIAL");  // server startup ?
    k = InStr(Caps(LoadedMapName), "MAPVOTEX");     // server startup ?
    if (j >= 0 || k >= 0)
    {
        return InitializeMvx();
    }
    else                                            // local botmatch
    if (Level.NetMode == NM_StandAlone)
    {
        sGame = Level.Game.Gamename;
        CheckForGameName(sGame);
        if (LoadedGameNum == -1)
        {
            sGame = string(Level.Game.Class);
            CheckForGameClass(sGame);
            if (LoadedGameNum == -1)
            {
                return InitializeMvx();
            }
        }
        SaveGameInfo(CustomGame[LoadedGameNum].GameName); // save to MVGameName
    }

    CheckServername();
    Settings = "";

	if ( DefaultSettings != "" )
	{
		Settings = DefaultSettings;
		if (Right(Settings, 1) != ",")
		    Settings = Settings$",";
        sSet = "";
		if ( Len(Settings) > 3 )
		{
 		    for (j=0; j<100; j++)                          // 100 default settings
		    {
			 	pos = InStr(Settings,",");
			 	if ( pos < 0 )
			 		break;
                sTemp = (Left(Settings, pos));
                sAly = sTemp;
                Settings = Mid(Settings,pos + 1);
                sTemp = MVGA.CheckForAlias(sTemp);
                if (InStr(sAly, ">") > 0)
                    DebugLog("DefaultSetting Alias: "$sAly$" = "$sTemp);
                sSet = sSet$sTemp;
                if (Right(sSet, 1) != ",")
                    sSet = sSet$",";
			}
            if (sSet != "")
                Settings = sSet;
            else
                Settings = DefaultSettings;

            Settings = RemoveRemarks(Settings);
            DebugLog("Default Settings = "$Settings);

            if ( Len(Settings) > 2 )
            {
                for (j=0; j<100; j++)                      // 100 default settings
                {
                    pos = InStr(Settings, ",");
                    if ( pos < 0 )
                        break;
                    else
                    {
				        ExecuteSetting(Left(Settings, pos), True);
				        Settings = Mid(Settings,pos + 1);
                    }
                }
            }
		}
	}

    class'ChallengeBotInfo'.default.bRandomOrder = bRandomBotOrder;

    bFound = false;
    Mutators = "";
    Settings = "";
    UnloadMutators = "";

    i = LoadedGameNum;
    if (CustomGame[i].bNoEndGameVoting)
        bNoEndGameVoting = True;

    bFound = true;
	if ( CustomGame[i].Mutators != "" )
	{
		Mutators = CustomGame[i].Mutators;
        if (Right(Mutators, 1) != ",")
            Mutators = Mutators$",";
        if ( Len(Mutators) > 2 )
        {
            sSet = "";
            for (j=0; j<100; j++)                  // 100 game mutators
            {
                pos = InStr(Mutators, ",");
                if ( pos < 0 )
                    break;
                sTemp = (Left(Mutators, pos));
                Mutators = Mid(Mutators,pos + 1);
                sAly = sTemp;
                x = InStr(sTemp, ";");
                if (x < 0)
                {
                    sTemp = MVGA.CheckForAlias(sTemp);
                    if (InStr(sAly, ">") > 0)
                        DebugLog("GameMutator Alias: "$sAly$" = "$sTemp);
                    sSet = sSet$sTemp$",";
                }
            }

            if (sSet != "")
            {
                Mutators = "";
                for (j=0; j<100; j++)              // game unloaded mutators
                {
                     k = InStr(sSet, ",");
                     if (k == -1)
                         break;

                     sTemp = Left(sSet, k);
                     sSet = Mid(sSet, k+1);

                     k = InStr(sTemp, "-");
                     if (k != -1)
                         UnloadMutators = UnloadMutators$Mid(sTemp, 1)$",";
                     else
                         Mutators = Mutators$sTemp$",";
                }
            }
            else
                Mutators = CustomGame[i].Mutators;

            Mutators = RemoveRemarks(Mutators);
            DebugLog("CustomGame Mutators = "$Mutators$" | UnloadMutators = "$UnloadMutators);

            if ( Len(Mutators) > 2 )
            {
               for (j=0; j<100; j++)               // 100 game mutators
               {
                   pos = InStr(Mutators, ",");
                   if ( pos < 0 )
                   {
                      break;
                   }
                   else
                   {
                      StartActorMutator(Left(Mutators, pos));
                      Mutators = Mid(Mutators,pos + 1);
                   }
               }
            }
        }
	}

	if ( CustomGame[i].Settings != "" )
	{
		Settings = CustomGame[i].Settings;
		if (Right(Settings, 1) != ",")
		    Settings = Settings$",";
        sSet = "";
		if ( Len(Settings) > 3 )
		{
 		    for (j=0; j<100; j++)                  // 100 CustomGame settings
		    {
			 	pos = InStr(Settings,",");
			 	if ( pos < 0 )
			 		break;
                sTemp = (Left(Settings, pos));
                sAly = sTemp;
                Settings = Mid(Settings,pos + 1);
                sTemp = MVGA.CheckForAlias(sTemp);
                if (InStr(sAly, ">") > 0)
                    DebugLog("GameSetting Alias: "$sAly$" = "$sTemp);
                sSet = sSet$sTemp;
                if (Right(sSet, 1) != ",")
                    sSet = sSet$",";
			}
            if (sSet != "")
                Settings = sSet;
            else
                Settings = CustomGame[i].Settings;

            Settings = RemoveRemarks(Settings);
            DebugLog("CustomGame Settings = "$Settings);

            if ( Len(Settings) > 2 )
            {
                for (j=0; j<100; j++)              // 100 CustomGame settings
                {
                    pos = InStr(Settings, ",");
                    if ( pos < 0 )
                        break;
                    else
                    {
				        ExecuteSetting(Left(Settings, pos), False);
				        Settings = Mid(Settings,pos + 1);
                    }
                }
            }
		}
	}

    Mutators = "";
    Settings = "";

    if ( DefaultMutators != "" )
    {
       Mutators = DefaultMutators;

       if (Right(Mutators, 1) != ",")
           Mutators = Mutators$",";
       sSet = "";
       if ( Len(Mutators) > 2 )
       {
          for (j=0; j<100; j++)                            // 100 default mutators
          {
              pos = InStr(Mutators, ",");
              if ( pos < 0 )
                  break;
              sTemp = (Left(Mutators, pos));
              Mutators = Mid(Mutators,pos + 1);
              x = InStr(sTemp, ";");
              if (x < 0)
              {
                  sAly = sTemp;
                  sTemp = MVGA.CheckForAlias(sTemp);
                  if (InStr(sAly, ">") > 0)
                      DebugLog("DefaultMutator Alias: "$sAly$" = "$sTemp);
                  sSet = sSet$sTemp;
                  if (Right(sSet, 1) != ",")
                      sSet = sSet$",";
              }
          }

          if (sSet != "")
              Mutators = sSet;
          else
              Mutators = DefaultMutators;

          Mutators = RemoveRemarks(Mutators);
          DebugLog("Default Mutators = "$Mutators);

          if ( Len(Mutators) > 2 )
          {
             for (j=0; j<100; j++)                         // 100 default mutators
             {
                 pos = InStr(Mutators, ",");
                 if ( pos < 0 )
                     break;
                 else
                 {
                    sTemp = Left(Mutators, pos);
                    StartActorMutator(sTemp);
                    Mutators = Mid(Mutators,pos + 1);
                 }
             }
          }
       }
    }

	return true;
}

function string GetInlineSettings(int gNum, string sGame) // ?FragLimit= etc
{
    local string sLine, sDeft, sTemp, sSet;
    local int j, k;

    sTemp = "";
    sDeft = "";

    if (sGame != "")
    {
        if (Left(sGame, 4) == "[GG]")
            sGame = Mid(sGame, 4);
        for (j=0; j<MaxNumGames; j++)
        {
             if (CustomGame[j].GameName ~= sGame)
             {
                 gNum = j;                                 // sets game num
                 break;
             }
        }
    }

    if (MVGA == None)
        MVGA = Spawn(class'MapVoteGameAliases');

	sDeft = DefaultSettings;
	if (sDeft != "")
	{
		if (Right(sDeft, 1) != ",")
		    sDeft = sDeft$",";
	}

	if ( CustomGame[gNum].Settings != "" )
	{
		sLine = CustomGame[gNum].Settings;
		if (Right(sLine, 1) != ",")
		    sLine = sLine$",";

		sLine = sDeft$sLine;
        sSet = "";
		if ( Len(sLine) > 3 )
		{
 		    for (j=0; j<200; j++)                          // 200 default + game settings
		    {
			 	pos = InStr(sLine,",");
			 	if ( pos < 0 )
			 		break;
                sTemp = (Left(sLine, pos));
                sLine = Mid(sLine,pos + 1);
                if (Left(sTemp, 1) != ";")
                {
                    sTemp = MVGA.CheckForAlias(sTemp);
                    sSet = sSet$sTemp;
                    if (Right(sSet, 1) != ",")
                        sSet = sSet$",";
                }
			}

            if (sSet != "")
                sLine = sSet;
            else
                sLine = CustomGame[gNum].Settings;

			sTemp = "";
			for (j=0; j<200; j++)                          // 200 default + game settings
			{
				k = InStr(sLine,",");
				if ( k < 0 )
				{
					break;
				}
				else
				{
				    sTemp = sTemp$"?"$Left(sLine, k);
					sLine = Mid(sLine, k+1);
				}
			}
		}
	}

	sLine = RemoveRemarks(sTemp);
	DebugLog("InLine Settings = "$sLine);
	return sLine;
}

function string RemoveRemarks(string sLine)
{
    local string sTemp, sNew;
    local int i, j;

    if (InStr(sLine, ";") == -1)       // no remarks found
        return sLine;

    if (Right(sLine, 1) != ",")
        sLine = sLine$",";

    sNew = "";

    for (i=0; i<100; i++)
    {
        j = InStr(sLine, ",");
        if (j == -1)
            break;
        sTemp = Left(sLine, j+1);      // include ,
        sLine = Mid(sLine, j+1);
        if (Left(sTemp, 1) != ";")
            sNew = sNew$sTemp;
    }

    DebugLog("MVX After Remarked Line = "$sNew);
    return sNew;
}

function StartActorMutator (string ClassString)
{
	local Actor A;
	local Class<Actor> ActorClass;

    if (Left(ClassString, 1) == ";" || Left(ClassString, 1) == "-" || Left(ClassString, 1) == "<") // remark, unload or alias
        return;

    if (InStr(Caps(UnloadMutators), Caps(ClassString)) != -1) // unload
    {
        DebugLog("Remove Mutator: "$ClassString);
        return;
    }

	DebugLog("Start  Mutator: "$ClassString);
	ActorClass=Class<Actor>(DynamicLoadObject(ClassString,Class'Class'));
	if ( ActorClass == None )
	{
		Log("Starting Actor/Mutator FAILED. Could not DynamicLoadObject:" @ ClassString,'MapVoteX');
		return;
	}
	foreach AllActors(ActorClass,A)
	{
		Log("Actor already spawned:" @ ClassString,'MapVoteX');
		return;
	}
	A=Spawn(ActorClass);
	if ( A == None )
	{
		Log("Spawning Actor/Mutator FAILED:" @ ClassString,'MapVoteX');
		return;
	}
	if ( A.IsA('Mutator') )
	{
		Level.Game.BaseMutator.AddMutator(Mutator(A));
	}
}

function ExecuteSetting(string Setting, bool bIsDefaultSetting)
{
	local string Property;
	local string Value;
	local int i;

	Property = Left(Setting,InStr(Setting, "=" ));
	Value = Mid(Setting,InStr(Setting, "=") + 1);

    i = InStr(Caps(Property), ";");
    if (i >= 0)
        return;
    i = InStr(Caps(Property), "<");
    if (i >= 0)
        return;
    i = InStr(Caps(Property), ">");
    if (i >= 0)
        return;

	i = int(Value);

    if (Caps(Property) == "TICKRATE")
    {
        Level.ConsoleCommand("Set ipDrv.TcpNetDriver NetServerMaxTickRate "$i);
        Level.ConsoleCommand("Set ipDrv.TcpNetDriver LanServerMaxTickRate "$i);
        Level.ConsoleCommand("Set Engine.DemoRecDriver NetServerMaxTickRate "$i);
        Level.ConsoleCommand("Set Engine.DemoRecDriver LanServerMaxTickRate "$i);
        TickRating = i;
    }
    else
    if (Caps(Property) == "GOALTEAMSCORE" || Caps(Property) == "FRAGLIMIT")
    {
        if (Level.Game.IsA('CTFGame'))
            CTFGame(Level.Game).GoalTeamScore = i;
        else
        if (Level.Game.IsA('Domination') || Level.Game.IsA('Assault'))
            Domination(Level.Game).GoalTeamScore = i;
        else
        if (Level.Game.IsA('TeamGamePlus') || Level.Game.bTeamGame)
        {
            TeamGamePlus(Level.Game).GoalTeamScore = i;
            DeathMatchPlus(Level.Game).FragLimit = i;
        }
        else
            DeathMatchPlus(Level.Game).FragLimit = i;
    }
    else
    if (Caps(Property) == "DIFFICULTY" || Caps(Property) == "SKILL")
    {
        Level.Game.Difficulty = i;
        DeathMatchPlus(Level.Game).Difficulty = i;
        class'ChallengeBotInfo'.default.Difficulty = i;
    }
    else
    if (Caps(Property) == "MAXTEAMS")
    {
        if (Level.Game.IsA('Assault'))
            class'Botpack.Assault'.default.MaxTeams = i;
        else
        if (Level.Game.IsA('Domination'))
        {
            class'Botpack.Domination'.default.MaxTeams = i;
            class'Botpack.TeamGamePlus'.default.MaxTeams = i;
        }
        else
        if (Level.Game.IsA('CTFGame'))
            class'Botpack.CTFGame'.default.MaxTeams = i;
        else
        if (Level.Game.IsA('TeamGamePlus'))
        {
            class'Botpack.TeamGamePlus'.default.MaxTeams = i;
        }
    }
    else
    if (Caps(Property) == "MAXTEAMSIZE")
    {
        if (Level.Game.IsA('Assault'))
            class'Botpack.Assault'.default.MaxTeamSize = i;
        else
        if (Level.Game.IsA('Domination'))
            class'Botpack.Domination'.default.MaxTeamSize = i;
        else
        if (Level.Game.IsA('CTFGame'))
            class'Botpack.CTFGame'.default.MaxTeamSize = i;
        else
        if (Level.Game.IsA('TeamGamePlus'))
            class'Botpack.TeamGamePlus'.default.MaxTeamSize = i;
    }
    else
    if (Caps(Property) == "BWEAPONSTAY")
    {
        if (Value ~= "True")
        {
            i = 1;
            DeathMatchPlus(Level.Game).bMultiWeaponStay = True;
            DeathMatchPlus(Level.Game).bCoopWeaponMode  = True;
        }
        else
        {
            i = 0;
            DeathMatchPlus(Level.Game).bMultiWeaponStay = False;
            DeathMatchPlus(Level.Game).bCoopWeaponMode  = False;
        }
    }
    else
    if (Caps(Property) == "MAXPLAYERS")
    {
        DeathMatchPlus(Level.Game).MaxPlayers = i;
        Level.Game.MaxPlayers = i;
    }
    else
    if (Caps(Property) == "BTOURNAMENT")
    {
        DeathMatchPlus(Level.Game).bTournament = bool(i);
    }
    else
    if (Caps(Property) == "BUSETRANSLOCATOR" || Caps(Property) == "BTRANSLOCATOR")
    {
        DeathMatchPlus(Level.Game).bUseTranslocator = bool(i);
    }

	Level.Game.SetPropertyText(Property, Value);

	if ( bIsDefaultSetting )
		Log("Execute Default Setting:" @ Setting $ " | Property:" @ Property $ " | Value:" @ Value,'MapVoteX');
	else
		Log("Execute Game Setting:" @ Setting $ " | Property:" @ Property $ " | Value:" @ Value,'MapVoteX');
}

function string GetAbsoluteTime()
{
	local string AbsoluteTime;

	if (Level.Hour < 10)
		AbsoluteTime = AbsoluteTime$"0"$Level.Hour;
	else
		AbsoluteTime = AbsoluteTime$Level.Hour;

	if (Level.Minute < 10)
		AbsoluteTime = AbsoluteTime$":0"$Level.Minute;
	else
		AbsoluteTime = AbsoluteTime$":"$Level.Minute;

	if (Level.Second < 10)
		AbsoluteTime = AbsoluteTime$":0"$Level.Second;
	else
		AbsoluteTime = AbsoluteTime$":"$Level.Second;

	return AbsoluteTime;
}

function SaveGameInfo(string sGame)
{
    MVGameName = sGame;
    SaveConfig();
}

function SavePackageActors(int iGame)
{
    local string sGame;

    sGame = CustomGame[iGame].GameClass;
    sGame = Caps(sGame);

    if (bEnableServerPackages && Level.Netmode != NM_StandAlone)
    {
        if (bClearAllPackageActors)
        {
            MVDF = Spawn(class'MapVoteDefaultPackages');
            if (MVDF.DefaultPackage[0] != "")
            {
                MVDF.ClearDefaultPackages();
                MVDF.RestoreDefaultPackages();
            }

            MVDA = Spawn(class'MapVoteDefaultActors');
            if (MVDA.DefaultActor[0] != "")
            {
                MVDA.ClearDefaultActors();
                MVDA.RestoreDefaultActors();
            }

            bClearAllPackageActors=False;
            SaveConfig();
        }

        MVSP = Spawn(class'MapVoteServerPackages');
        MVSP.MyMut = self;
        MVSP.gmPackages = CustomGame[iGame].ServerPackages;
        MVSP.WriteServerPackages(iGame);

        MVSA = Spawn(class'MapVoteServerActors');
        MVSA.MyMut = self;
        MVSA.gmActors = CustomGame[iGame].ServerActors;
        MVSA.WriteServerActors(iGame);
    }
}

function SpawnAdvancedClasses()
{
    if (!bClassesSpawned)
    {
        MVCS = Spawn(class'MapVoteClientSplash');
        MVCW = Spawn(class'MapVoteClientWelcome');
        MVWL = Spawn(class'MapVoteWelcomeLinks');
        MVAM = Spawn(class'MapVoteAdminMessages');
        MVCM = Spawn(class'MapVoteCrashedMaps');
        MVSP = Spawn(class'MapVoteServerPackages');
        MVSA = Spawn(class'MapVoteServerActors');
        bClassesSpawned = true;
    }
}

function string GetTimeSeconds()
{
    local string sTime;

    sTime = "";
	if (Level.Second < 10)
		sTime = sTime$"0"$Level.Second;
	else
		sTime = sTime$Level.Second;

    return sTime;
}

defaultproperties
{
    bAutoOpen=True
    bAllowMidGameVoting=True
    bRandomBotOrder=True
    bEnableSplashLogo=True
    bEnableWelcomeMenu=True
    bEmptyServerSwitchMap=True
    bEmptyServerSwitch=True
    bSortCustomGameList=True
    bDoubleClickVoting=True
    bQuickGameVote=True
    bShowChatCommands=True
    bUseServerNames=True
    bAddToFavorites=True
    bAllowPlayerKillBots=True
    bAllowPlayerAddBot=True
    NoEndGameVotingDelay=40
    EmptyServerSwitchTimeMins=15
    VoteTimeLimit=60
    StartMesgDelay=8
    PopupWindowDelay=20
    RepeatLimit=5
    GameStartVoteDelay=15
    DefaultMutators=""
    DefaultSettings="MaxTeams=2"
    DefaultCustomGameNum=0
    CustomGame(0)=(bUse=True,MapFolder="Maps/DM",DefaultMap="DM-Deck16][",GameName="Tournament DeathMatch",GameClass="Botpack.DeathMatchPlus",Mutators="",Settings="InitialBots=3,Difficulty=4,Fraglimit=20,TimeLimit=20,MinPlayers=4",ServerPackages="",ServerActors="",bNoEndGameVoting=False)
    CustomGame(1)=(bUse=True,MapFolder="Maps/TDM,Maps/DM",DefaultMap="DM-Deck16][",GameName="Tournament Team Game",GameClass="Botpack.TeamGamePlus",Mutators="",Settings="InitialBots=5,Fraglimit=30,TimeLimit=20,MinPlayers=6",ServerPackages="",ServerActors="",bNoEndGameVoting=False)
    CustomGame(2)=(bUse=True,MapFolder="Maps/CTF",DefaultMap="CTF-Face",GameName="Capture the Flag",GameClass="Botpack.CTFGame",Mutators="",Settings="InitialBots=5,GoalTeamScore=5,TimeLimit=20,MinPlayers=6",ServerPackages="",ServerActors="",bNoEndGameVoting=False)
    CustomGame(3)=(bUse=True,MapFolder="Maps/DOM",DefaultMap="DOM-Olden",GameName="Domination",GameClass="Botpack.Domination",Mutators="",Settings="InitialBots=7",ServerPackages="",ServerActors="",bNoEndGameVoting=False)
    CustomGame(4)=(bUse=True,MapFolder="Maps/AS",DefaultMap="AS-Frigate",GameName="Assault",GameClass="Botpack.Assault",Mutators="",Settings="InitialBots=7",ServerPackages="",ServerActors="",bNoEndGameVoting=False)
    CustomGame(5)=(bUse=True,MapFolder="Maps/DM",DefaultMap="DM-Deck16][",GameName="Last Man Standing",GameClass="Botpack.DeathMatchPlus",Mutators="",Settings="",ServerPackages="",ServerActors="",bNoEndGameVoting=False)
    Version="112"
}

